from enum import Enum


class BeamValue(str, Enum):
    """Type of beam behaviour at a note"""
    # https://www.w3.org/2021/06/musicxml40/musicxml-reference/data-types/beam-value/
    backward_hook = "backward hook"
    begin = "begin"
    continue_beam = "continue"
    end = "end"
    forward_hook = "forward hook"



import os
from pathlib import Path


# cache location
# (inspired by https://github.com/huggingface/datasets)
DEFAULT_XDG_CACHE_HOME = "~/.cache"
XDG_CACHE_HOME = os.getenv("XDG_CACHE_HOME", DEFAULT_XDG_CACHE_HOME)
DEFAULT_MC_CACHE_HOME = os.path.join(XDG_CACHE_HOME, "smashcima")
MC_CACHE_HOME = os.path.expanduser(os.getenv("MC_HOME", DEFAULT_MC_CACHE_HOME))

# default asset repository path
DEFAULT_MC_ASSETS_CACHE = os.path.join(MC_CACHE_HOME, "assets")
MC_ASSETS_CACHE = Path(os.getenv("MC_ASSETS_CACHE", DEFAULT_MC_ASSETS_CACHE))



import abc
from typing import Set

from smashcima.geometry import Point, Transform
from smashcima.scene import AffineSpace, Glyph


class GlyphSynthesizer(abc.ABC):
    """Interface for a glyph synthesizer"""

    @abc.abstractmethod
    def supports_label(self, label: str) -> bool:
        """Returns true if the given glyph type can be synthesized.
        
        Override this when building custom glyph synthesizers.
        """
        raise NotImplementedError

    def synthesize_glyph(
        self,
        label: str,
        parent_space: AffineSpace,
        transform: Transform
    ) -> Glyph:
        """Synthesizes a new glyph
        
        :param label: Which kind of glyph to synthesize (classification label)
        :param parent_space: The glyph will be placed into this space
        :param transform: How should the glyph be placed into the parent space
        :returns: The newly synthesized glyph instance, already attached to
            the parent space.
        """
        # check that we are not accidentally passing in an Enum instead of str
        assert type(label) is str, "The requested glyph label is not a string"

        # create a new glyph
        glyph = self.create_glyph(label)

        # check its label
        assert glyph.label == label, f"The user requested class {label} " + \
            f"but got {glyph.label} instead"
        
        # attach to the parent space
        glyph.space.parent_space = parent_space
        glyph.space.transform = transform

        return glyph

    def synthesize_glyph_at(
        self,
        label: str,
        parent_space: AffineSpace,
        point: Point
    ):
        """Synthesizes a new glyph with its origin translated to a point.

        :param label: Which kind of glyh to synthesize (classification label)
        :param parent_space: The glyph will be placed into this space
        :param point: The glyph's origin will be placed over this point.
            The coordinates are relative to the parent space.
        """
        return self.synthesize_glyph(
            label=label,
            parent_space=parent_space,
            transform=Transform.translate(point.vector)
        )
    
    def create_glyph(self, label: str) -> Glyph:
        """Creates a new glyph instance of the specified label.

        Override this method when building custom glyph synthesizers.
        The returned glyph should have no parent space.

        :param label: The kind of glyph to create.
        """
        raise NotImplementedError



from .Point import Point
from .Rectangle import Rectangle
from .Quad import Quad
from typing import List
import numpy as np


class Polygon:
    """
    Polygon is a list of points that enclose an area.
    """

    def __init__(self, points: List[Point]):
        self.points = points

    @staticmethod
    def from_rectangle(rectangle: Rectangle) -> "Polygon":
        """Constructs a polygon from a rectangle by going over its corners
        from top left, in the clocwise direction."""
        return Polygon([
            rectangle.top_left_corner,
            rectangle.top_right_corner,
            rectangle.bottom_right_corner,
            rectangle.bottom_left_corner
        ])

    @staticmethod
    def from_quad(quad: Quad) -> "Polygon":
        """Constructs a polygon from a quad."""
        return Polygon([quad.a, quad.b, quad.c, quad.d])

    @staticmethod
    def from_cv2_contour(contour: np.ndarray) -> "Polygon":
        """Constructs a polygon from an OpenCV contour instance"""
        # enumerated points are vertical vectors: [[X, Y]]
        return Polygon([Point(p[0, 0], p[0, 1]) for p in contour])

    def __repr__(self):
        return f"Quad({self.a}, {self.b}, {self.c}, {self.d})"

    def bbox(self) -> Rectangle:
        """Returns the bounding box of the polygon"""
        pts = self.points
        top = min(p.y for p in pts)
        bottom = max(p.y for p in pts)
        left = min(p.x for p in pts)
        right = max(p.x for p in pts)
        return Rectangle(
            x=left,
            y=top,
            width=right-left,
            height=bottom-top
        )



from typing import Tuple, List
import bisect


class Skyline:
    """Stacks boxes by dropping them from the top, building a skyline and
    determining for each next box where to place it."""
    def __init__(self, ground_level: float = 0):
        self.pins: List[Tuple[float, float]] = [
            (float("-Inf"), ground_level)
        ]
        """Going from left to right, a pin is a point (X, Y) that defines
        the new level ground of the skyline from that point to the left
        until the next pin is hit."""

    def drop_box(
        self,
        minimum: float,
        maximum: float,
        thickness: float
    ) -> float:
        """Minimum and maximum are the sideways bounds of the box and
        thickness is the "height" of the box falling from the top.
        Returned value is the top-position of the placed box."""
        assert minimum < maximum
        assert thickness > 0

        start_index = bisect.bisect_left(self.pins, minimum, key=lambda p: p[0])
        end_index = bisect.bisect_left(self.pins, maximum, key=lambda p: p[0])

        tallest_level = max(p[1] for p in self.pins[start_index-1:end_index])
        new_top = tallest_level + thickness

        new_pins = self.pins[:start_index] + [
            (minimum, new_top),
            (maximum, self.pins[end_index-1][1])
        ] + self.pins[end_index:]

        self.pins = new_pins
        self.compress_pins()
        return new_top
    
    def overlay_box(
        self,
        minimum: float,
        maximum: float,
        level: float
    ):
        """Places a box on the ground and updates the skyline appropriately"""
        assert minimum < maximum

        start_index = bisect.bisect_left(self.pins, minimum, key=lambda p: p[0])
        end_index = bisect.bisect_left(self.pins, maximum, key=lambda p: p[0])

        new_pins = self.pins[:start_index]

        if new_pins[-1][1] <= level:
            new_pins.append((minimum, level))

        for i in range(start_index, end_index):
            pin = self.pins[i]
            if pin[1] > level:
                new_pins.append(pin)
            else:
                new_pins.append((pin[0], level))
        
        new_pins.append(
            (maximum, self.pins[end_index-1][1])
        )
        
        new_pins += self.pins[end_index:]

        self.pins = new_pins
        self.compress_pins()
    
    def compress_pins(self):
        new_pins = [self.pins[0]]

        for pin in self.pins[1:]:
            last_position, last_level = new_pins[-1]
            
            # same level, skip
            if last_level == pin[1]:
                continue

            # same position, remove last
            if last_position == pin[0]:
                new_pins.pop()
            
            new_pins.append(pin)

        self.pins = new_pins


# .venv/bin/python3 -m smashcima.synthesis.layout.column.Skyline
if __name__ == "__main__":
    # drop boxes
    s = Skyline()
    assert s.drop_box(0, 2, thickness=1) == 1
    assert s.pins[1:] == [(0, 1), (2, 0)]

    #
    #OO
    ###########

    assert s.drop_box(1, 4, thickness=1) == 2
    assert s.pins[1:] == [(0, 1), (1, 2), (4, 0)]

    # OOO
    #OOxx
    ###########

    assert s.drop_box(0, 2, thickness=1) == 3
    assert s.pins[1:] == [(0, 3), (2, 2), (4, 0)]

    #OO
    #xOOO
    #OOxx
    ###########

    # overlay boxes
    s = Skyline()
    s.overlay_box(0, 2, level=1)
    assert s.pins[1:] == [(0, 1), (2, 0)]

    #
    #OO
    ###########

    s.overlay_box(1, 4, level=2)
    assert s.pins[1:] == [(0, 1), (1, 2), (4, 0)]

    # OOO
    #xOOO
    ###########

    s.overlay_box(2, 6, level=1)
    assert s.pins[1:] == [(0, 1), (1, 2), (4, 1), (6, 0)]

    # xxx
    #xxOOOO
    ###########

    s.overlay_box(0, 6, level=2)
    assert s.pins[1:] == [(0, 2), (6, 0)]

    #OOOOOO
    #OOOOOO
    ###########



from .Point import Point
from .Rectangle import Rectangle
from typing import List


class Quad:
    """
    Quad is a set of four points enclosing a quadrangle area.
    This is what you get from a rectangle after doing an affine transform.
    You can turn it back to a rectangle by getting the bounding box (bbox).
    It's created from a rectangle by going over its corners from the left top
    corner in the clockwise direction.
    """

    def __init__(self, a: Point, b: Point, c: Point, d: Point):
        self.a = a
        self.b = b
        self.c = c
        self.d = d

    @staticmethod
    def from_rectangle(rectangle: Rectangle) -> "Quad":
        """Constructs a quad from a rectangle by going over its corners
        from top left, in the clocwise direction."""
        return Quad(
            a=rectangle.top_left_corner,
            b=rectangle.top_right_corner,
            c=rectangle.bottom_right_corner,
            d=rectangle.bottom_left_corner
        )

    @property
    def points(self) -> List[Point]:
        """Returns the quad points as a list"""
        return [self.a, self.b, self.c, self.d]
    
    def __repr__(self):
        return f"Quad({self.a}, {self.b}, {self.c}, {self.d})"
    
    def bbox(self) -> Rectangle:
        """Returns the bounding box of the quad"""
        pts = self.points
        top = min(p.y for p in pts)
        bottom = max(p.y for p in pts)
        left = min(p.x for p in pts)
        right = max(p.x for p in pts)
        return Rectangle(
            x=left,
            y=top,
            width=right-left,
            height=bottom-top
        )



from typing import List
from smashcima.geometry.Point import Point
import numpy as np
import cv2


# Inspired by:
# https://stackoverflow.com/questions/67143809/finding-the-end-points-of-a-hand-drawn-line-with-opencv


def get_line_endpoints(mask: np.ndarray) -> List[Point]:
    """Runs an algorithm to find all endpoints of lines in the given mask.
    The input mask is a true/false (1/0) mask with uint8 type from MPP
    crop object."""
    assert len(mask.shape) == 2
    assert mask.dtype == np.uint8

    # pad the mask with one pixel wide margin of zeros
    # (this makes sure skeletonization does not break at the edges)
    # (it also resolves the 2D convolution border problems)
    mask = cv2.copyMakeBorder(mask, 1, 1, 1, 1, cv2.BORDER_CONSTANT, value=0)

    # get the skeleton (image of 0s and 1s)
    skeleton = cv2.ximgproc.thinning(mask * 255) // 255
    
    # count the number of neighbors for each pixel (using a convolution)
    kernel = np.ones(shape=(3, 3), dtype=np.uint8)
    kernel[1, 1] = 0
    neighbor_count = cv2.filter2D(
        src=skeleton,
        ddepth=-1,
        kernel=kernel,
    )
    neighbor_count[skeleton != 1] = 0 # only count neighbors for skeleton pixels

    # get only the pixels where the neighbor count is 1
    neighbor_count[neighbor_count != 1] = 0

    # enumerate these points
    cv_points = cv2.findNonZero(neighbor_count)

    if cv_points is None:
        return []

    return [Point(x, y) for x, y in cv_points.squeeze(axis=1)]



import requests
import tqdm
from pathlib import Path


def download_file(url: str, path: Path, with_progress_bar=True):
    """Downloads a file with a progress bar and saves it to a path"""
    if with_progress_bar:
        print("Downloading " + str(url))
        print("and saving it to " + str(path))
    
    # taken from:
    # https://stackoverflow.com/questions/37573483/progress-bar-while-download-file-over-http-with-requests/37573701
    response = requests.get(url, stream=True)
    total_size_in_bytes = int(response.headers.get("Content-Length", 0))
    block_size = 1024
    progress_bar = tqdm.tqdm(
        total=total_size_in_bytes,
        unit="iB",
        unit_scale=True,
        disable=not with_progress_bar
    )
    with open(path, "wb") as file:
        for data in response.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)
    progress_bar.close()
    if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
        raise Exception("File downloading failed.")



from .Contours import Contours
from .Point import Point
from .Polygon import Polygon
from .Quad import Quad
from .Rectangle import Rectangle
from .Transform import Transform
from .Vector2 import Vector2

from .units import mm_to_px, px_to_mm



from smashcima.scene.visual.Page import Page
from smashcima.scene.visual.StaffVisual import StaffVisual
from smashcima.geometry.Vector2 import Vector2
from smashcima.geometry.Rectangle import Rectangle
from smashcima.geometry.Transform import Transform
from smashcima.scene.ViewBox import ViewBox
from smashcima.scene.AffineSpace import AffineSpace
from .StafflinesSynthesizer import StafflinesSynthesizer
from .PaperSynthesizer import PaperSynthesizer
from typing import List
from dataclasses import dataclass


@dataclass
class PageSetup:
    size: Vector2

    padding_top: float
    padding_bottom: float
    padding_left: float
    padding_right: float

    staff_count: int


HANDWRITTEN_PAGE_SETUP = PageSetup(
    # inspired by:
    # https://www.digitalniknihovna.cz/mzk/view/uuid:378147a5-3d3f-47f0-9f31-80740052e86c?page=uuid:74c8e808-1dd3-4ff7-bea5-658b7d4cb1a5
    size = Vector2(250, 350),
    padding_top=20,
    padding_bottom=40,
    padding_left=10,
    padding_right=10,
    staff_count=13
)


MUSESCORE_PAGE_SETUP = PageSetup(
    # A4 page with MuseScore-like padding
    size = Vector2(210, 297),
    padding_top=10,
    padding_bottom=10,
    padding_left=10,
    padding_right=10,
    staff_count=12
)


class SimplePageSynthesizer:
    """Synthesizes page layout"""
    def __init__(
        self,
        stafflines_synthesizer: StafflinesSynthesizer,
        paper_synthesizer: PaperSynthesizer
    ):
        self.stafflines_synthesizer = stafflines_synthesizer
        self.paper_synthesizer = paper_synthesizer

        self.page_setup = MUSESCORE_PAGE_SETUP

    def synthesize_page(self, page_origin: Vector2) -> Page:
        page_space = AffineSpace(
            transform=Transform.translate(page_origin)
        )

        view_box = ViewBox(
            rectangle=Rectangle(
                page_origin.x,
                page_origin.y,
                self.page_setup.size.x,
                self.page_setup.size.y
            ),
            space=page_space
        )

        self.paper_synthesizer.synthesize_paper(
            page_space=page_space,
            placement=Rectangle(
                0, 0,
                self.page_setup.size.x,
                self.page_setup.size.y
            )
        )

        staves = self._synthesize_stafflines(page_space)

        return Page(
            space=page_space,
            view_box=view_box,
            staves=staves
        )
    
    def _synthesize_stafflines(self, page_space: AffineSpace) -> List[StaffVisual]:
        content_box = Rectangle(
            self.page_setup.padding_left,
            self.page_setup.padding_top,
            self.page_setup.size.x \
                - (self.page_setup.padding_left + self.page_setup.padding_right),
            self.page_setup.size.y \
                - (self.page_setup.padding_top + self.page_setup.padding_bottom)
        )

        starting_point = content_box.y \
            + self.stafflines_synthesizer.staff_height / 2

        vertical_step = (content_box.height \
            - self.stafflines_synthesizer.staff_height) \
            / (self.page_setup.staff_count - 1)

        staves: List[StaffVisual] = []
        for i in range(self.page_setup.staff_count):
            stafflines = self.stafflines_synthesizer.synthesize_stafflines(
                page_space=page_space,
                position=Vector2(
                    content_box.x,
                    starting_point + i * vertical_step
                ),
                width=content_box.width
            )
            staves.append(stafflines)
        
        return staves



from .MuscimaPPGlyphs import MuscimaPPGlyphs
from ...AssetRepository import AssetRepository
from pathlib import Path
import sys


# NOTE: This code re-installs the MuscimaPPGlyphs asset bundle.
# It cannot be part of that file in if __name__ == "__main__" idiom,
# because it would cause a double-import, potentially leading to issues:
# https://stackoverflow.com/questions/43393764/python-3-6-project-structure-leads-to-runtimewarning

assets = AssetRepository(Path("smashcima_assets"))

print("Re-installing MUSCIMA++ glyphs...")
bundle = assets.resolve_bundle(MuscimaPPGlyphs, force_install=True)

if len(sys.argv) >= 2 and sys.argv[1] == "--debug":
    print("Building the debug folder...")
    bundle.build_debug_folder()

print("Done.")



from smashcima.scene.semantic.Score import Score
from smashcima.scene.semantic.ScoreEvent import ScoreEvent
from smashcima.scene.visual.StaffVisual import StaffVisual
from smashcima.synthesis.GlyphSynthesizer import GlyphSynthesizer
from smashcima.synthesis.LineSynthesizer import LineSynthesizer
from .NoteheadsColumn import NoteheadsColumn, synthesize_noteheads_column
from .RestsColumn import RestsColumn, synthesize_rests_column
from .ColumnBase import ColumnBase
from .AugmentationDotsColumn import AugmentationDotsColumn, \
    synthesize_augmentation_dots_column
from .AccidentalsColumn import AccidentalsColumn, synthesize_accidentals_column
from typing import List
import random


class EventColumn(
    NoteheadsColumn,
    RestsColumn,
    AugmentationDotsColumn,
    AccidentalsColumn
):
    def __init__(self, *args, **kwargs):
        # ignore constructors of parents, initialize only the super-parent
        ColumnBase.__init__(self, *args, **kwargs)
    
    def __post_init__(self):
        NoteheadsColumn.__post_init__(self)
        RestsColumn.__post_init__(self)
        AugmentationDotsColumn.__post_init__(self)
        AccidentalsColumn.__post_init__(self)

    def _position_glyphs(self):
        NoteheadsColumn._position_glyphs(self)
        RestsColumn._position_glyphs(self)
        AugmentationDotsColumn._position_glyphs(self)
        AccidentalsColumn._position_glyphs(self)


def synthesize_event_column(
    staves: List[StaffVisual],
    rng: random.Random,
    glyph_synthesizer: GlyphSynthesizer,
    line_synthesizer: LineSynthesizer,
    score: Score,
    score_event: ScoreEvent
) -> EventColumn:
    column = EventColumn(staves, rng.random())

    synthesize_noteheads_column(
        column=column,
        staves=staves,
        glyph_synthesizer=glyph_synthesizer,
        line_synthesizer=line_synthesizer,
        score=score,
        score_event=score_event
    )

    synthesize_rests_column(
        column=column,
        staves=staves,
        glyph_synthesizer=glyph_synthesizer,
        line_synthesizer=line_synthesizer,
        score=score,
        score_event=score_event,
        rng=rng
    )

    synthesize_augmentation_dots_column(
        column=column,
        staves=staves,
        glyph_synthesizer=glyph_synthesizer,
        score=score,
        score_event=score_event
    )

    synthesize_accidentals_column(
        column=column,
        staves=staves,
        glyph_synthesizer=glyph_synthesizer,
        score=score,
        score_event=score_event
    )
    
    return column



from dataclasses import dataclass
from ..SceneObject import SceneObject
from fractions import Fraction
from .TypeDuration import TypeDuration


@dataclass
class Durable(SceneObject):
    """Durable is something that has duration (note or rest)"""
    
    type_duration: TypeDuration
    """The duration type that the note has visually (e.g. an eighth note)"""
    
    augmentation_dots: int
    """How many augmentation (duration) dots are there
    to augment the note's duration"""
    
    fractional_duration: Fraction
    """How many quarter notes does the note actually take up (e.g. an eighth
    note triplet takes up less time than a regular eighth note)"""



from smashcima.scene.AffineSpace import AffineSpace
from smashcima.geometry.Rectangle import Rectangle
import abc


class PaperSynthesizer(abc.ABC):
    """
    Represents an interface to a synthesizer that produces sprites
    with background paper texture
    """

    @abc.abstractmethod
    def synthesize_paper(
        self,
        page_space: AffineSpace,
        placement: Rectangle
    ):
        """Synthesizes a new paper texture

        :page_space The affine space of the page we are synthesizing
        :placement Where inside the page space should the paper be placed
        """
        raise NotImplementedError



import punq
from typing import TypeVar, Type


T = TypeVar("T")
U = TypeVar("U", bound=T)


class Container:
    """
    Service container used for Model services. Since it's meant to be used
    in the context of a single model, it by default registers all services
    as singletons, since it's unlikely there would be required any transients.
    """

    def __init__(self, register_itself=True):
        self._container = punq.Container()

        # register the container itself into the container
        if register_itself:
            self.instance(Container, self)
    
    def instance(self, instance_type: Type[T], instance: T):
        self._container.register(
            service=instance_type,
            instance=instance
        )

    def type(self, concrete_type: Type[T]):
        self._container.register(
            service=concrete_type,
            scope=punq.Scope.singleton
        )

    def interface(self, abstract_type: Type[T], concrete_type: Type[U]):
        self._container.register(
            service=abstract_type,
            factory=concrete_type,
            scope=punq.Scope.singleton
        )
    
    def resolve(self, resolve_type: Type[T]) -> T:
        return self._container.resolve(resolve_type)



from dataclasses import dataclass
from typing import Optional
from .Durable import Durable
from .Pitch import Pitch
from .AccidentalValue import AccidentalValue


@dataclass
class Note(Durable):
    pitch: Pitch
    "Musical pitch of the note"

    accidental_value: Optional[AccidentalValue]
    "What accidental is attached to the note"



from dataclasses import dataclass
from .Measure import Measure
from .ScoreEvent import ScoreEvent
from .StaffSemantic import StaffSemantic
from typing import List, Iterator, Tuple


@dataclass
class ScoreMeasure:
    """A view of an entire score in a time-wise manner"""
    
    measures: List[Measure]
    "Part-measures ordered just like parts in the score"

    events: List[ScoreEvent]
    "Temporally-ordered musical events (sets of durables with the same onset)"

    @staticmethod
    def from_part_measures(measures: List[Measure]) -> "ScoreMeasure":
        return ScoreMeasure(
            measures=measures,
            events=ScoreEvent.merge_from_measures(measures)
        )
    
    def measure_for_staff(self, staff_index: int) -> Measure:
        """Returns the part measure that contains the given staff index"""
        assert staff_index < sum(len(m.staves) for m in self.measures), \
            "Staff index is larger than the number of staves in the score"
        mi = 0
        skipped_staves = 0
        while skipped_staves + len(self.measures[mi].staves) < staff_index:
            mi += 1
            skipped_staves += len(self.measures[mi].staves)
        return self.measures[mi]

    def iterate_staves_with_measures(
        self
    ) -> Iterator[Tuple[StaffSemantic, Measure]]:
        """Iterates over semantic staves and their measures in the score"""
        for measure in self.measures:
            for staff in measure.staves:
                yield (staff, measure)



from dataclasses import dataclass

from ..LabeledRegion import LabeledRegion
from ..SceneObject import SceneObject
from ..semantic.Measure import Measure
from ..semantic.StaffSemantic import StaffSemantic
from ..SmashcimaLabels import SmashcimaLabels
from .StaffVisual import StaffVisual


@dataclass
class StaffMeasure(SceneObject):
    """Region in the score representing a measure on one staff"""

    measure: Measure
    """Semantic measure this visual staff measure represents a part of"""

    staff_semantic: StaffSemantic
    """Which semantic staff of the measure do we represent"""

    staff_visual: StaffVisual
    """Which visual staff does this staff measure sit on"""

    region: LabeledRegion
    """The region encapsulating the staff measure"""

    def __post_init__(self):
        assert self.region.label == SmashcimaLabels.staffMeasure.value, \
            "Staff measure region must have proper classification label"

    @classmethod
    def of_staff_semantic(cls, staff_semantic: StaffSemantic):
        return cls.of(staff_semantic, lambda sm: sm.staff_semantic)

    @classmethod
    def many_of_staff_visual(cls, staff_visual: StaffVisual):
        return cls.many_of(staff_visual, lambda sm: sm.staff_visual)



from smashcima.geometry.Contours import Contours
from smashcima.geometry.Point import Point
from smashcima.geometry.Polygon import Polygon
from smashcima.geometry.Quad import Quad
from smashcima.scene.Glyph import Glyph
from smashcima.scene.LabeledRegion import LabeledRegion
from smashcima.scene.SmashcimaLabels import SmashcimaLabels
from smashcima.scene.visual.Flag import Flag
from smashcima.scene.visual.StaffMeasure import StaffMeasure
from smashcima.scene.visual.StaffVisual import StaffVisual
from smashcima.scene.semantic.Score import Score
from smashcima.scene.visual.System import System
from smashcima.scene.visual.Page import Page
from smashcima.scene.visual.Stem import Stem
from smashcima.scene.AffineSpace import AffineSpace
from smashcima.scene.semantic.ScoreMeasure import ScoreMeasure
from smashcima.scene.semantic.BeamedGroup import BeamedGroup
from smashcima.scene.semantic.Chord import Chord
from smashcima.scene.semantic.Note import Note
from smashcima.scene.semantic.TypeDuration import TypeDuration
from smashcima.geometry.Transform import Transform
from smashcima.scene.SmuflLabels import SmuflLabels
from smashcima.scene.visual.SystemMeasure import SystemMeasure
from smashcima.synthesis.GlyphSynthesizer import GlyphSynthesizer
from ..BeamStemSynthesizer import BeamStemSynthesizer
from ...LineSynthesizer import LineSynthesizer
from .Column import Column
from .ColumnBase import ColumnBase
from .BarlinesColumn import synthesize_barlines_column
from .ClefsColumn import synthesize_header_clefs
from .EventColumn import synthesize_event_column
from typing import List, Optional, Dict
import random


class _SystemState:
    """Holds state while synthesizing a single system of music"""

    def __init__(self, minimal_column_spacing: float):
        self.minimal_column_spacing = minimal_column_spacing
        "Minimum spacing inserted in between columns"
        
        self.columns: List[Column] = []
        "List of columns in this system"

        self.total_width = 0.0
        "Total width of all the columns, if stacked most tightly"

        self.current_measure_index: Optional[int] = None
        "What measure are we currently in? None means no measure (header/footer)"

        self.measure_indices: List[int] = []
        "List of measures in the system, indicated by their score index"

        self.columns_per_measure: Dict[int, List[Column]] = {}
        "Columns attached to their measure by score measure index"
    
    def enter_measure(self, measure_index: int):
        assert measure_index not in self.measure_indices, \
            "You cannot enter a measure twice"
        self.current_measure_index = measure_index
        self.measure_indices.append(measure_index)
    
    def exit_measure(self):
        self.current_measure_index = None
    
    @property
    def measure_count(self) -> int:
        """How many measures are present in the system"""
        return len(self.measure_indices)
    
    def append_column(self, column: Column):
        column.position_glyphs()
        
        if len(self.columns) > 0:
            self.total_width += self.minimal_column_spacing

        self.total_width += column.width
        self.columns.append(column)

        if self.current_measure_index is not None:
            self.columns_per_measure.setdefault(self.current_measure_index, [])
            self.columns_per_measure[self.current_measure_index].append(column)
    
    def delete_measure(self, measure_index: int):
        assert measure_index in self.measure_indices
        assert measure_index in self.columns_per_measure

        columns = self.columns_per_measure[measure_index]
        del self.columns_per_measure[measure_index]
        self.measure_indices.remove(measure_index)

        for column in columns:
            self.total_width -= column.width + self.minimal_column_spacing
            self.columns.remove(column)
            column.detach()


def _place_columns_tightly(state: _SystemState, minimal_column_spacing: float):
    """Places columns from left to right with minimal spacing between them"""
    time_position = 0
    for column in state.columns:
        column.time_position = time_position + column.left_width
        column.position_glyphs()
        time_position += column.width + minimal_column_spacing


def _place_columns_flexbox(
    state: _SystemState,
    minimal_column_spacing: float,
    available_width: float
):
    """Places columns stretched out to fill the staff width, like CSS flexbox"""
    # can be negative when overflowing
    width_to_distribute = available_width - state.total_width
    if width_to_distribute >= 0:
        total_flex = sum(c.flex_grow for c in state.columns)
    else:
        total_flex = sum(c.flex_shrink for c in state.columns)
    width_unit = width_to_distribute / total_flex

    time_position = 0
    for column in state.columns:
        width_portion = width_unit * column.flex_grow
        time_position += width_portion / 2
        
        column.time_position = time_position + column.left_width
        column.position_glyphs()
        time_position += column.width + minimal_column_spacing
        
        time_position += width_portion / 2


# TODO: define a layout synthesizer interface and inherit
# IN -> semantic scene object graph
# OUT -> visual scene object graph
class ColumnLayoutSynthesizer:
    def __init__(
        self,
        glyph_synthesizer: GlyphSynthesizer,
        line_synthesizer: LineSynthesizer,
        beam_stem_synthesizer: BeamStemSynthesizer,
        rng: random.Random
    ):
        self.glyph_synthesizer = glyph_synthesizer
        self.line_synthesizer = line_synthesizer
        self.beam_stem_synthesizer = beam_stem_synthesizer
        self.rng = rng

        # The following values control the synthesizer.
        # The page-flowing options are set to respect explicit breaks
        # and ignore overflowing. This guarantees the same visual layout
        # as the one encoded in the input MusicXML.

        self.minimal_column_spacing = 1.2
        "Minimal spacing inserted between columns"

        self.stretch_out_columns = True
        "Whether to place columns tightly or to stretch them out to fill staff"

        self.place_debug_rectangles = False
        "Places column rectangles for debugging (to see the columns)"

        self.respect_line_and_page_breaks = True
        "If true, the layout respects the page formatting data of the score"

        self.disable_wrapping = True
        "When true, only linebreaks and page breaks control flow"

    def fill_page(
        self,
        page: Page,
        score: Score,
        start_on_measure: int
    ) -> List[System]:
        """Fills the page with music and returns the list of synthesized systems"""
        systems: List[System] = []
        
        # state
        remaining_staves = len(page.staves)
        completed_staves = 0
        current_measure = start_on_measure

        # synthesize systems until we have the space available
        while remaining_staves >= score.staff_count:

            # or until there are measures left to synthesize
            if current_measure >= score.measure_count:
                break

            # or until we hit a page break
            if self.respect_line_and_page_breaks \
                    and not current_measure == start_on_measure:
                if current_measure in score.new_page_measure_indices:
                    break

            # synthesize a single system
            system = self.synthesize_system(
                page_space=page.space,
                staves=page.staves[
                    completed_staves:
                    completed_staves+score.staff_count
                ],
                score=score,
                start_on_measure=current_measure
            )
            systems.append(system)

            # update state
            remaining_staves -= score.staff_count
            completed_staves += score.staff_count
            current_measure += system.measure_count
        
        return systems
    
    def synthesize_system(
        self,
        page_space: AffineSpace,
        staves: List[StaffVisual],
        score: Score,
        start_on_measure: int
    ) -> System:
        """Synthesizes a single system of music onto the provided staves"""
        assert start_on_measure < score.measure_count, \
            "There must be at least one measure left to be placed on the system"

        assert len(staves) == score.staff_count, \
            "Given staves do not match the required number of staves per system"

        for staff_visual in staves:
            assert staff_visual.space.parent_space is page_space, \
                "Given staves do not live in the given page space"

        # === phase 1: synthesizing columns ===

        state = _SystemState(
            minimal_column_spacing=self.minimal_column_spacing
        )
        available_width = min(sf.width for sf in staves)

        # synthesize header = start of the system signatures
        # TODO: extract this into a method
        state.append_column(
            synthesize_header_clefs(
                staves, self.rng, self.glyph_synthesizer,
                score, start_on_measure
            )
        )

        # synthesize measures...
        next_measure_index = start_on_measure
        while True:

            # until there is space left
            if not self.disable_wrapping:
                if state.total_width > available_width:
                    break

            # or until there are measures to synthesize
            if next_measure_index >= score.measure_count:
                break

            # or until we hit a line break
            if self.respect_line_and_page_breaks \
                    and not next_measure_index == start_on_measure:
                if next_measure_index in score.new_system_measure_indices \
                        or next_measure_index in score.new_page_measure_indices:
                    break

            state.enter_measure(next_measure_index)
            score_measure = score.get_score_measure(next_measure_index)
            next_measure_index += 1

            # TODO: extract this into a method, there's going to be a lot
            # of code here eventually (synthesize_measure)

            # construct a column for each event
            for score_event in score_measure.events:
                state.append_column(
                    synthesize_event_column(
                        staves, self.rng, self.glyph_synthesizer,
                        self.line_synthesizer, score, score_event
                    )
                )
            
            # column for the barlines
            state.append_column(
                synthesize_barlines_column(
                    staves, self.rng, self.glyph_synthesizer
                )
            )

            state.exit_measure()

        # TODO: synthesize system footer (key and time changes)

        # remove measures from the end until we stop overflowing
        # (and we want at least one measure to reman no matter what)
        # (and only do that if we have wrapping enabled)
        if not self.disable_wrapping:
            while state.total_width >= available_width and state.measure_count > 1:
                state.delete_measure(state.measure_indices[-1])

        # === phase 2: placing columns ===

        # tight or flexbox stretch
        if self.stretch_out_columns:
            _place_columns_flexbox(
                state,
                self.minimal_column_spacing,
                available_width
            )
        else:
            _place_columns_tightly(state, self.minimal_column_spacing)

        # optionally place debugging boxes around columns
        if self.place_debug_rectangles:
            for column in state.columns:
                if isinstance(column, ColumnBase):
                    column.place_debug_boxes()

        # === phase 3: construct the system object ===

        # TODO: create layout bboxes and add them to the system
        system = System(
            # TODO: this scene object should be implemented fully
            # (like the SystemMeasure is)
            first_measure_index=start_on_measure,
            measure_count=state.measure_count
        )

        # NOTE: this is a very rough measure boundary,
        # this needs to be refactored before being used for training
        # (i.e. leading barline and staff header are not included now)
        for measure_index, columns in state.columns_per_measure.items():
            start_time = min(c.time_position - c.left_width for c in columns)
            end_time = max(c.time_position + c.right_width for c in columns)

            score_measure = score.get_score_measure(measure_index)
            staff_measures: List[StaffMeasure] = []

            for staff_visual, (staff_semantic, measure) in zip(
                staves,
                score_measure.iterate_staves_with_measures(),
                strict=True
            ):
                a = staff_visual.staff_coordinate_system.get_transform(
                    pitch_position=4, time_position=start_time
                ).apply_to(Point(0, 0))
                b = staff_visual.staff_coordinate_system.get_transform(
                    pitch_position=4, time_position=end_time
                ).apply_to(Point(0, 0))
                c = staff_visual.staff_coordinate_system.get_transform(
                    pitch_position=-4, time_position=end_time
                ).apply_to(Point(0, 0))
                d = staff_visual.staff_coordinate_system.get_transform(
                    pitch_position=-4, time_position=start_time
                ).apply_to(Point(0, 0))
                contours = Contours([Polygon([a, b, c, d])])

                staff_measures.append(StaffMeasure(
                    measure=measure,
                    staff_semantic=staff_semantic,
                    staff_visual=staff_visual,
                    region=LabeledRegion(
                        space=staff_visual.space,
                        contours=contours,
                        label=SmashcimaLabels.staffMeasure.value
                    )
                ))
            
            # construct the complete system measure
            polygons: List[Polygon] = []
            for sm in staff_measures:
                polygons += sm.region.get_contours_in_space(page_space).polygons
            SystemMeasure(
                score_measure=score_measure,
                staff_measures=staff_measures,
                region=LabeledRegion(
                    space=page_space,
                    contours=Contours(polygons),
                    label=SmashcimaLabels.systemMeasure.value
                )
            )

        # === phase 4: synthesizing beams, stems and flags ===

        for i in range(system.measure_count):
            score_measure = score.get_score_measure(
                system.first_measure_index + i
            )
            self.beam_stem_synthesizer.synthesize_beams_and_stems_for_measure(
                page_space,
                score_measure
            )
            self.synthesize_flags_in_measure(
                page_space,
                score_measure
            )
        
        # === phase 5: replacing ligatures ===

        # TODO: ligatures
        # (this should probbably be done somewhere else than here)

        return system
    
    def synthesize_flags_in_measure(
        self,
        page_space: AffineSpace,
        score_measure: ScoreMeasure
    ):
        # get all chords
        chords: List[Chord] = []
        for score_event in score_measure.events:
            for event in score_event.events:
                for durable in event.durables:
                    if isinstance(durable, Note):
                        chord = Chord.of_note(durable)
                        if chord is not None and chord not in chords:
                            chords.append(chord)
        
        # go through the chords and add flags
        for chord in chords:
            stem = Stem.of_chord_or_none(chord)
            if stem is None:
                continue

            if len(chord.notes) == 0:
                continue
            
            if BeamedGroup.of_chord_or_none(chord) is not None:
                continue

            type_duration=chord.notes[0].type_duration
            
            if type_duration.to_quarter_multiple() > \
                TypeDuration.eighth.to_quarter_multiple():
                continue

            # get the glyph class
            label = SmuflLabels.flag_from_type_duration_and_stem_value(
                type_duration=type_duration,
                stem_value=chord.stem_value
            ).value

            # create the glyph
            glyph = self.glyph_synthesizer.synthesize_glyph_at(
                label=label,
                parent_space=page_space,
                point=stem.tip.transform_to(page_space)
            )
            Flag(
                glyph=glyph,
                stem=stem
            )



from typing import List, Sequence

import numpy as np

from .Polygon import Polygon
from .Rectangle import Rectangle


class Contours:
    """Contours is a list of polygons that together encapsulate a shape"""

    def __init__(self, polygons: List[Polygon]):
        self.polygons = polygons

    @staticmethod
    def from_cv2_contours(cv_contours: Sequence[np.ndarray]) -> "Contours":
        """Constructs a polygon from an OpenCV contour instance"""
        return Contours([
            Polygon.from_cv2_contour(cv_contour)
            for cv_contour in cv_contours
        ])

    def bbox(self) -> Rectangle:
        """Returns the bounding box of all contours"""
        point_cloud = [
            point for polygon in self.polygons for point in polygon.points
        ]
        return Polygon(point_cloud).bbox()



from dataclasses import dataclass, field
from ..SceneObject import SceneObject
from typing import List
from .Durable import Durable


@dataclass
class StaffSemantic(SceneObject):
    """
    Represents a staff within a measure (e.g. piano measures have 2 staves),
    as a collection of durables
    """

    staff_number: int
    "Number of this staff, e.g. 1, 2, 3. Numbered from 1 and from the top down"

    durables: List[Durable] = field(default_factory=list)
    "Links to all durables within this staff"

    @property
    def staff_index(self) -> int:
        """Zero-based index of the staff in measure stafflines"""
        return self.staff_number - 1

    @classmethod
    def of_durable(cls, durable: Durable):
        return cls.of(durable, lambda s: s.durables)



from ..SceneObject import SceneObject
from ..LineGlyph import LineGlyph
from .Notehead import Notehead
from .RestVisual import RestVisual
from dataclasses import dataclass, field
from typing import List, Optional


@dataclass
class LedgerLine(SceneObject):
    """Ledger line (visual), extending the stafflines outside their range"""
    
    glyph: LineGlyph

    affected_noteheads: List[Notehead] = field(default_factory=list)
    """List of noteheads that are affected by this ledger line (so if we are
    at the top of the staff, all notes that are on or above this ledger line
    fall into this list; notes that are below this ledger line are not affected.
    The list contains only noteheads from the same onset event.)
    If this list is empty, then this ledger line affects a (whole/half) rest."""

    affected_rest: Optional[RestVisual] = None
    """A ledger line is also needed for whole and half rests. Here, the
    placement is simpler, because there are no intermediate lines drawn,
    only the one that directly touches the rest."""

    def detach(self):
        """Unlink the glyph from the scene"""
        self.glyph.detach()
        self.affected_noteheads = None
        self.affected_rest = None



from smashcima.scene import SmashcimaLabels
from smashcima.scene.semantic.Score import Score
from smashcima.scene.semantic.Event import Event
from smashcima.scene.semantic.StaffSemantic import StaffSemantic
from smashcima.scene.semantic.ScoreEvent import ScoreEvent
from smashcima.scene.semantic.RestSemantic import RestSemantic
from smashcima.scene.visual.StaffVisual import StaffVisual
from smashcima.scene.visual.RestVisual import RestVisual
from smashcima.scene.visual.LedgerLine import LedgerLine
from smashcima.scene.SmuflLabels import SmuflLabels
from smashcima.synthesis.GlyphSynthesizer import GlyphSynthesizer
from smashcima.synthesis.LineSynthesizer import LineSynthesizer
from smashcima.geometry.Point import Point
from smashcima.random_between import random_between
from .ColumnBase import ColumnBase
from typing import List
import random


class RestsColumn(ColumnBase):
    def __post_init__(self) -> None:
        self.rests: List[RestVisual] = []

    def add_rest(self, rest_glyph: RestVisual):
        assert rest_glyph.rest_semantic is not None
        self.glyphs.append(rest_glyph.glyph)
        self.rests.append(rest_glyph)
    
    def _position_glyphs(self):
        self.position_rests()
    
    def position_rests(self):
        for rest in self.rests:
            display_pitch = rest.rest_semantic.display_pitch \
                or RestVisual.default_display_pitch(
                    rest.clef, rest.rest_semantic.type_duration
                )
            pitch_position = RestVisual.display_pitch_to_glyph_pitch_position(
                rest.clef, display_pitch, rest.rest_semantic.type_duration
            )

            rest.glyph.space.transform = rest.staff.staff_coordinate_system \
                .get_transform(
                    pitch_position=pitch_position,
                    time_position=self.time_position
                )


def synthesize_rests_column(
    column: RestsColumn,
    staves: List[StaffVisual],
    glyph_synthesizer: GlyphSynthesizer,
    line_synthesizer: LineSynthesizer,
    score: Score,
    score_event: ScoreEvent,
    rng: random.Random
):
    # for all the rests (including measure rests)
    for event in score_event.events:
        for durable in event.durables:
            if not isinstance(durable, RestSemantic): # inlcudes MeasureRest
                continue
            
            # resolve context
            event = Event.of_durable(durable)
            staff_sem = StaffSemantic.of_durable(durable)
            clef = event.attributes.clefs[staff_sem.staff_number]
            staff_index = score.staff_index_of_durable(durable)
            staff = staves[staff_index]

            # resolve pitch position
            display_pitch = durable.display_pitch \
                or RestVisual.default_display_pitch(
                    clef, durable.type_duration
                )
            pitch_position = RestVisual.display_pitch_to_glyph_pitch_position(
                clef, display_pitch, durable.type_duration
            )

            # create the rest
            label = SmuflLabels.rest_from_type_duration(
                durable.type_duration
            )
            glyph = glyph_synthesizer.synthesize_glyph_at(
                label=label.value,
                parent_space=staff.space,
                point=Point(0, 0) # glyph positioned later
            )
            rest = RestVisual(
                glyph=glyph,
                rest_semantic=durable,
                clef=clef,
                staff=staff,
                pitch_position=pitch_position,
            )
            column.add_rest(rest)

            # create ledger line for whole/half rests
            # (and attach it under the glyph space for simplicity)
            _synthesize_ledger_line_if_necessary(
                rest=rest,
                label=label,
                line_synthesizer=line_synthesizer,
                rng=rng
            )


def _synthesize_ledger_line_if_necessary(
    rest: RestVisual,
    label: SmuflLabels,
    line_synthesizer: LineSynthesizer,
    rng: random.Random
):
    # the rest is not whole nor half, no ledger line needed
    if label not in [SmuflLabels.restWhole, SmuflLabels.restHalf]:
        return

    # the rest is still within the staff, no ledgerline needed
    if abs(rest.pitch_position) < 4:
        return

    width = rest.glyph.get_bbox_in_space(rest.glyph.space).width \
        * random_between(1.2, 2.5, rng)
    
    glyph = line_synthesizer.synthesize_line(
        label=SmashcimaLabels.ledgerLine.value,
        parent_space=rest.glyph.space,
        start_point=Point(-width / 2, 0),
        end_point=Point(width / 2, 0)
    )
    LedgerLine(
        glyph=glyph,
        affected_noteheads=[], # none
        affected_rest=rest
    )



from dataclasses import dataclass

from smashcima.geometry import Contours, Rectangle

from .AffineSpace import AffineSpace
from .SceneObject import SceneObject


@dataclass
class Region(SceneObject):
    """A region with an area inside the scene.

    A region can be transformed through the affine space hierarchy
    and can be used for various area-describing tasks (object localization,
    masking, etc.).

    It is defined by a set of polygons, because a polygon can undergo affine
    transformations and remain a polygon (unlike a bounding box).
    Other lower-detail regions can be derived from the set of polygons easily
    (e.g. convex hull, tight rectangle, bounding box).
    """

    space: AffineSpace
    """Space in which the region is situated. Not owned by the region."""

    contours: Contours
    """Polygon areas that define the region (in parent space coordinates)"""

    def detach(self):
        """Unlink the region from the scene"""
        self.space = None

    @classmethod
    def many_of_space(cls, space: AffineSpace):
        return cls.many_of(space, lambda r: r.space)

    def get_contours_in_space(self, target_space: AffineSpace) -> Contours:
        """Returns the contours polygons transformed to the target space
        
        :param target_space: The space to which coordinates of the contours
            should be transformed. Must be an ancestor of this region's space.
        """
        transform = target_space.transform_from(self.space)
        return transform.apply_to(self.contours)

    def get_bbox_in_space(self, target_space: AffineSpace) -> Rectangle:
        """Returns the bounding box rectangle in the target space coordinates
        
        :param target_space: The space to which coordinates of the contours
            should be transformed. Must be an ancestor of this region's space.
        """
        return self.get_contours_in_space(target_space).bbox()



import base64
import xml.etree.ElementTree as ET
from typing import Optional

import cv2

from smashcima.geometry import Transform
from smashcima.scene import (AffineSpace, AffineSpaceVisitor, LabeledRegion,
                             SceneObject, Sprite, ViewBox)

SVG_NS = "{http://www.w3.org/2000/svg}"
XLINK_NS = "{http://www.w3.org/1999/xlink}"


class SvgExporter:
    def __init__(
        self,
        background_fill: Optional[str] = None,
        render_labeled_regions: bool = False
    ):
        self.background_fill = background_fill
        self.render_labeled_regions = render_labeled_regions

    def export_string(self, view_box: ViewBox, pretty=False) -> str:
        """Exports the provided scene into an SVG string"""
        
        # build the element tree
        root = self.export(view_box)

        # prettify
        if pretty and hasattr(ET, "indent"):
            ET.indent(root, " ", 0)
        elif pretty:
            print(
                "WARNING: SVG pretty-printing is only supported " +
                "from Python 3.9 and up."
            )

        # register namespaces and stringify
        ET.register_namespace(
            prefix="",
            uri="http://www.w3.org/2000/svg"
        )
        ET.register_namespace(
            prefix="xlink",
            uri="http://www.w3.org/1999/xlink"
        )
        return str(ET.tostring(
            root,
            encoding="utf-8",
            xml_declaration=True
        ), "utf-8")

    def export(self, view_box: ViewBox) -> ET.Element:
        """Exports the provided scene into and SVG XML element tree"""
        
        # build the root SVG element
        root_element = ET.Element(SVG_NS + "svg")
        width = view_box.rectangle.width
        height = view_box.rectangle.height
        root_element.attrib = {
            "width": f"{width}mm",
            "height": f"{height}mm",
            "viewBox": f"0 0 {width} {height}",
            "version": "1.1",
        }

        # background fill color rectangle
        if self.background_fill is not None:
            background_element = ET.Element(SVG_NS + "rect")
            background_element.attrib = {
                "x": "0",
                "y": "0",
                "width": str(width),
                "height": str(height),
                "fill": self.background_fill,
                "id": "SvgBackgroundFill"
            }
            root_element.append(background_element)

        # get the root space and viewport transform
        root_space = view_box.space.get_root()
        root_to_view_transform = root_space.transform_from(view_box.space) \
            .inverse()
        
        # extract visible object hierarchy
        svg_visitor = SvgVisitor(space=root_space)
        svg_visitor.run()

        # set the viewport transofrm to the root space group
        # (i.e. position the whole scene so that it aligns with the viewport)
        svg_visitor.group_element.attrib = {
            "transform": svg_matrix_from_transform(root_to_view_transform),
            "id": "RootAffineSpace_" + str(id(root_space))
        }
        root_element.append(svg_visitor.group_element)

        # extract labeled regions overlay
        if self.render_labeled_regions:
            regions_svg_group = ET.Element(SVG_NS + "g")
            regions_svg_group.attrib = {
                "id": "LabeledRegion overlay"
            }
            region_visitor = RegionVisitor(
                space=root_space,
                regions_svg_group=regions_svg_group,
                transform_to_viewport=root_to_view_transform
            )
            region_visitor.run()
            root_element.append(regions_svg_group)

        return root_element


class SvgVisitor(AffineSpaceVisitor):
    """Visits an AffineSpace hierarchy and constructs an SVG tree"""
    
    def __init__(self, space: AffineSpace):
        super().__init__(space)

        self.group_element = ET.Element(SVG_NS + "g")
        self.group_element.attrib = {
            "transform": svg_matrix_from_transform(space.transform),
            "id": "AffineSpace_" + str(id(space))
        }
    
    def create_sub_visitor(self, sub_space: AffineSpace) -> "SvgVisitor":
        return SvgVisitor(sub_space)

    def accept_sub_visitor(self, sub_visitor: "SvgVisitor"):
        self.group_element.append(sub_visitor.group_element)

    def visit_scene_object(self, obj: SceneObject):
        if isinstance(obj, Sprite):
            self.group_element.append(
                sprite_to_image_element(sprite=obj)
            )
        # TODO: elif SVG elements in the scene


class RegionVisitor(AffineSpaceVisitor):
    """Constructs flattened SVG elements for labeled regions (because if they
    are part of the hierarchy, they will get rotated and transformed - we want
    them to be in the viewport space)"""

    def __init__(
        self,
        space: AffineSpace,
        regions_svg_group: ET.Element,
        transform_to_viewport: Transform
    ):
        super().__init__(space)

        self.regions_svg_group = regions_svg_group
        self.transform_to_viewport = transform_to_viewport
    
    def create_sub_visitor(self, sub_space: AffineSpace) -> "RegionVisitor":
        return RegionVisitor(
            space=sub_space,
            regions_svg_group=self.regions_svg_group,
            transform_to_viewport=sub_space.transform.then(
                self.transform_to_viewport
            )
        )

    def accept_sub_visitor(self, sub_visitor: "RegionVisitor"):
        pass # nothing needs to be done

    def visit_scene_object(self, obj: SceneObject):
        if isinstance(obj, LabeledRegion):
            self.regions_svg_group.append(
                labeled_region_to_svg_elements(
                    region=obj,
                    transform_to_viewport=self.transform_to_viewport
                )
            )


def svg_matrix_from_transform(t: Transform) -> str:
    """Formats a smashcima transform into the SVG transform attribute value"""
    a, b = t.matrix[:, 0]
    c, d = t.matrix[:, 1]
    e, f = t.matrix[:, 2]
    return f"matrix({a} {b} {c} {d} {e} {f})"


def sprite_to_image_element(sprite: Sprite) -> ET.Element:
    """Converts a Sprite instance to an SVG image element"""
    image_element = ET.Element(SVG_NS + "image")
    
    png_binary_data = cv2.imencode(".png", sprite.bitmap)[1].tobytes()
    base64_str_data = str(base64.b64encode(png_binary_data), "utf-8")

    width = sprite.physical_width
    height = sprite.physical_height

    image_element.attrib = {
        "transform": svg_matrix_from_transform(sprite.transform),
        "x": str(-width * sprite.bitmap_origin.x),
        "y": str(-height * sprite.bitmap_origin.y),
        "width": str(width),
        "height": str(height),
        XLINK_NS + "href": "data:image/png;base64," + base64_str_data,
        "id": "Sprite_" + str(id(sprite))
    }

    return image_element


def labeled_region_to_svg_elements(
    region: LabeledRegion,
    transform_to_viewport: Transform
) -> ET.Element:
    rectangle_element = ET.Element(SVG_NS + "rect")

    bbox = transform_to_viewport.apply_to(region.contours).bbox()

    # get a pseudo-random hue for the label
    hue = hash(region.label) % 360

    rectangle_element.attrib = {
        "x": str(bbox.x),
        "y": str(bbox.y),
        "width": str(bbox.width),
        "height": str(bbox.height),
        "fill": f"hsl({hue}, 100%, 50%)",
        "fill-opacity": "0.25",
        "stroke": f"hsl({hue}, 100%, 50%)",
        "stroke-opacity": "0.5",
        "stroke-width": "0.5",
        "id": "Region_" + str(id(region)) + "__" + region.label
    }

    return rectangle_element



from typing import Set, Dict
from ..GlyphSynthesizer import GlyphSynthesizer
from smashcima.scene.Glyph import Glyph
from smashcima.assets.AssetRepository import AssetRepository
from smashcima.assets.glyphs.muscima_pp.MuscimaPPGlyphs import MuscimaPPGlyphs
from ...scene.SmuflLabels import SmuflLabels
from smashcima.synthesis.style.MuscimaPPStyleDomain import MuscimaPPStyleDomain
import random


_QUERY_TO_MPP_LOOKUP: Dict[str, str] = {
    # barlines
    SmuflLabels.barlineSingle.value: SmuflLabels.barlineSingle.value,

    # clefs    (clefs ignore the normal/small distinction)
    SmuflLabels.gClef.value: SmuflLabels.gClef.value,
    SmuflLabels.gClefSmall.value: SmuflLabels.gClef.value,
    SmuflLabels.fClef.value: SmuflLabels.fClef.value,
    SmuflLabels.fClefSmall.value: SmuflLabels.fClef.value,
    SmuflLabels.cClef.value: SmuflLabels.cClef.value,
    SmuflLabels.cClefSmall.value: SmuflLabels.cClef.value,

    # noteheads
    SmuflLabels.noteheadWhole.value: SmuflLabels.noteheadWhole.value,
    SmuflLabels.noteheadHalf.value: SmuflLabels.noteheadWhole.value,
    SmuflLabels.noteheadBlack.value: SmuflLabels.noteheadBlack.value,

    # augmentation dot
    SmuflLabels.augmentationDot.value:SmuflLabels.augmentationDot.value,

    # flags
    SmuflLabels.flag8thUp.value: SmuflLabels.flag8thUp.value,
    SmuflLabels.flag8thDown.value: SmuflLabels.flag8thDown.value,
    SmuflLabels.flag16thUp.value: SmuflLabels.flag16thUp.value,
    SmuflLabels.flag16thDown.value: SmuflLabels.flag16thDown.value,

    # accidentals
    SmuflLabels.accidentalFlat.value: SmuflLabels.accidentalFlat.value,
    SmuflLabels.accidentalNatural.value: SmuflLabels.accidentalNatural.value,
    SmuflLabels.accidentalSharp.value: SmuflLabels.accidentalSharp.value,

    # rests
    SmuflLabels.restWhole.value: SmuflLabels.restWhole.value,
    SmuflLabels.restHalf.value: SmuflLabels.restHalf.value,
    SmuflLabels.restQuarter.value: SmuflLabels.restQuarter.value,
    SmuflLabels.rest8th.value: SmuflLabels.rest8th.value,
    SmuflLabels.rest16th.value: SmuflLabels.rest16th.value,
}


class MuscimaPPGlyphSynthesizer(GlyphSynthesizer):
    """Synthesizes glyphs by sampling from the MUSCIMA++ dataset"""
    
    def __init__(
        self,
        assets: AssetRepository,
        mpp_style_domain: MuscimaPPStyleDomain,
        rng: random.Random,
    ):
        bundle = assets.resolve_bundle(MuscimaPPGlyphs)
        self.symbol_repository = bundle.load_symbol_repository()
        "The symbol repository used for synthesis"

        self.mpp_style_domain = mpp_style_domain
        "Dictates which MUSCIMA++ writer to use for synthesis"
        
        self.rng = rng
        "RNG used for randomization"

    def supports_label(self, label: str) -> bool:
        return label in _QUERY_TO_MPP_LOOKUP

    def create_glyph(self, label: str) -> Glyph:
        # pick a glyph from the symbol repository
        if label in _QUERY_TO_MPP_LOOKUP:
            glyph = self.pick(_QUERY_TO_MPP_LOOKUP[label])
        else:
            raise Exception("Unsupported glyph label: " + label)

        # adjust its glyph class to match what the user wants
        # (because the mapping dictionary is not really 1:1)
        glyph.label = label

        return glyph
    
    def pick(self, label: str) -> Glyph:
        """Picks a random glyph from the symbol repository according to the
        current writer"""
        # get the list of glyphs to choose from
        # (if writer is missing this class, fall back on all writers)
        packed_glyphs = self.symbol_repository.glyphs_by_class_and_writer.get(
            (label, self.mpp_style_domain.current_writer)
        ) or self.symbol_repository.glyphs_by_class.get(label)

        if packed_glyphs is None or len(packed_glyphs) == 0:
            raise Exception(
                f"The glyph class {label} is not present in " + \
                "the symbol repository"
            )
        
        # pick a random glyph from the list
        packed_glyph = self.rng.choice(packed_glyphs)
        
        # deserialization here makes sure we create a new instance
        return packed_glyph.unpack()



import math
from typing import Iterator


class Vector2:
    """Mathematical 2D vector"""

    def __init__(self, x: float, y: float):
        self.x = float(x)
        self.y = float(y)
    
    @property
    def left(self) -> float:
        return self.x

    @property
    def top(self) -> float:
        return self.y

    @property
    def magnitude_squared(self) -> float:
        return self.x * self.x + self.y * self.y
    
    @property
    def magnitude(self) -> float:
        return math.sqrt(self.magnitude_squared)
    
    def __repr__(self):
        return f"Vector2({self.x}, {self.y})"

    def __add__(self, other):
        if isinstance(other, Vector2):
            return Vector2(self.x + other.x, self.y + other.y)
        else:
            raise ValueError("Vectors can only be added with other vectors")
    
    def __sub__(self, other):
        if isinstance(other, Vector2):
            return Vector2(self.x - other.x, self.y - other.y)
        else:
            raise ValueError("Vectors can only be subtracted from other vectors")
    
    def __neg__(self):
        return Vector2(-self.x, -self.y)

    def __mul__(self, other):
        if isinstance(other, float) or isinstance(other, int):
            return Vector2(self.x * other, self.y * other)
        else:
            raise ValueError("Vectors can only be multiplied by scalars")
    
    def __truediv__(self, other):
        if isinstance(other, float) or isinstance(other, int):
            return Vector2(self.x / other, self.y / other)
        else:
            raise ValueError("Vectors can only be divided by scalars")
    
    def __iter__(self) -> Iterator[float]:
        yield self.x
        yield self.y

    def rotate90degCC(self) -> "Vector2":
        """Returns a 90deg rotated version of the vector counter clockwise"""
        return Vector2(self.y, -self.x)

    def rotate90degCW(self) -> "Vector2":
        """Returns a 90deg rotated version of the vector clockwise"""
        return Vector2(-self.y, self.x)

    def normalize(self) -> "Vector2":
        """Returns the vector normalized to unit length"""
        m = self.magnitude
        if m == 0:
            raise Exception("Zero vector cannot be normalized")
        return self / m



from smashcima.scene.semantic.Score import Score
from smashcima.scene.semantic.ScoreEvent import ScoreEvent
from smashcima.scene.semantic.Note import Note
from smashcima.scene.semantic.RestSemantic import RestSemantic
from smashcima.scene.semantic.Durable import Durable
from smashcima.scene.visual.StaffVisual import StaffVisual
from smashcima.scene.visual.AugmentationDot import AugmentationDot
from smashcima.scene.visual.Notehead import Notehead
from smashcima.scene.visual.RestVisual import RestVisual
from smashcima.scene.SmuflLabels import SmuflLabels
from smashcima.synthesis.GlyphSynthesizer import GlyphSynthesizer
from smashcima.geometry.Point import Point
from smashcima.random_between import random_between
from .ColumnBase import ColumnBase
from typing import List, Dict, Optional, Union


class AugmentationDotsColumn(ColumnBase):
    def __post_init__(self) -> None:
        self.augmentation_dots: List[AugmentationDot] = []
    
    def add_augmentation_dot(self, augmentation_dot: AugmentationDot):
        assert len(augmentation_dot.owners) > 0
        self.glyphs.append(augmentation_dot.glyph)
        self.augmentation_dots.append(augmentation_dot)

    def _position_glyphs(self):
        self.place_augmentation_dots()

    def place_augmentation_dots(self):
        for staff in self.staves:
            self.place_augmentation_dots_for_staff(staff)

    def place_augmentation_dots_for_staff(self, staff: StaffVisual):
        noteheads: List[Notehead] = []
        notehead_dots: List[AugmentationDot] = []
        rests: List[RestVisual] = []
        rest_dots: List[AugmentationDot] = []
        for dot in self.augmentation_dots:
            for owner in dot.owners:
                if self.get_staff_of_glyph(owner.glyph) is not staff:
                    continue # skip glyphs in other staves
                if isinstance(owner, Notehead):
                    noteheads.append(owner)
                    notehead_dots.append(dot)
                elif isinstance(owner, RestVisual):
                    rests.append(owner)
                    rest_dots.append(dot)
        
        # where is the column in the staff space coords
        origin_x = staff.staff_coordinate_system.get_transform(
            0, self.time_position
        ).apply_to(Point(0, 0)).x

        # get right edge of all noteheads
        noteheads_right_edge = origin_x
        for notehead in noteheads:
            r = notehead.glyph.get_bbox_in_space(staff.space).right
            noteheads_right_edge = max(noteheads_right_edge, r)
        
        # constants
        BASE_OFFSET = random_between(1.0, 2.0, self.rng) # only used for noteheads
        SPACING = random_between(0.8, 1.6, self.rng) # between dots

        # place notehead dots
        for dot in notehead_dots:
            position_x = BASE_OFFSET + SPACING * (dot.augmentation_dot_index - 1)
            time_position = self.time_position + (
                noteheads_right_edge - origin_x
            ) + position_x
            dot.glyph.space.transform = staff.staff_coordinate_system.get_transform(
                pitch_position=dot.pitch_position,
                time_position=time_position
            )
        
        # place rest dots
        for dot in rest_dots:
            rest_right_edge = dot.owners[0].glyph.get_bbox_in_space(
                staff.space
            ).right
            position_x = SPACING * dot.augmentation_dot_index
            time_position = self.time_position + (
                rest_right_edge - origin_x
            ) + position_x
            dot.glyph.space.transform = staff.staff_coordinate_system.get_transform(
                pitch_position=dot.pitch_position,
                time_position=time_position
            )

def synthesize_augmentation_dots_column(
    column: AugmentationDotsColumn,
    staves: List[StaffVisual],
    glyph_synthesizer: GlyphSynthesizer,
    score: Score,
    score_event: ScoreEvent
):
    # cluster durables by staff
    durables: Dict[int, List[Durable]] = {}
    for event in score_event.events:
        for durable in event.durables:
            staff_index = score.staff_index_of_durable(durable)
            durables.setdefault(staff_index, [])
            durables[staff_index].append(durable)

    # for each staff durables
    for staff_index, staffline_durables in durables.items():
        created_dots: Dict[int, List[AugmentationDot]] = dict() # by pitch pos
        handled_noteheads: List[Notehead] = []

        # sort by pitch from high-to-low
        # (in order for noteheads to properly distribute dots)
        sorted_durables = list(staffline_durables)
        sorted_durables.sort(key=lambda d:
            d.pitch.get_linear_pitch() if isinstance(d, Note) else 0,
            reverse=True
        )

        for durable in sorted_durables:
            owner: Union[None, Notehead, RestVisual] = None
            pitch_position: Optional[int] = None
            augmentation_dot_count: int = 0

            # handle noteheads
            if isinstance(durable, Note):
                notehead = Notehead.of_note(durable)
                if notehead in handled_noteheads:
                    continue
                handled_noteheads.append(notehead)

                owner = notehead
                pitch_position = notehead.pitch_position
                augmentation_dot_count = durable.augmentation_dots

            # handle rests
            elif isinstance(durable, RestSemantic):
                rest_glyph = RestVisual.of_rest(durable)
                owner = rest_glyph
                pitch_position = rest_glyph.pitch_position
                augmentation_dot_count = durable.augmentation_dots

            # skip other durables (there should be none)
            else:
                continue

            # skip noteheads with no dots
            if augmentation_dot_count == 0:
                continue

            # alter pitch position if we are placed on line
            if pitch_position % 2 == 0:
                # try moving up if not taken
                if (pitch_position + 1) not in created_dots:
                    pitch_position += 1
                # else down
                elif (pitch_position - 1) not in created_dots:
                    pitch_position -= 1
                # else just up and reuse the dots there
                else:
                    pitch_position += 1

            # for all duration dots
            created_dots.setdefault(pitch_position, [])
            dots: List[AugmentationDot] = created_dots[pitch_position]
            for dot_index in range(1, augmentation_dot_count + 1):

                # reuse existing dot
                if dot_index <= len(dots):
                    dot = dots[dot_index - 1]
                    assert dot.augmentation_dot_index == dot_index
                    assert dot.pitch_position == pitch_position
                    dot.owners = [*dot.owners, owner]

                # create new dot
                glyph = glyph_synthesizer.synthesize_glyph_at(
                    label=SmuflLabels.augmentationDot.value,
                    parent_space=staves[staff_index].space,
                    point=Point(0, 0) # glyph positioned later
                )
                dot = AugmentationDot(
                    glyph=glyph,
                    owners=[owner],
                    augmentation_dot_index=dot_index,
                    pitch_position=pitch_position,
                )
                dots.append(dot)

                column.add_augmentation_dot(dot)



from smashcima.geometry.Point import Point
from smashcima.scene.visual.StaffVisual import StaffVisual
from smashcima.scene.Glyph import Glyph
from smashcima.scene.SmuflLabels import SmuflLabels
from smashcima.synthesis.GlyphSynthesizer import GlyphSynthesizer
from .ColumnBase import ColumnBase
from typing import List
import random


class BarlinesColumn(ColumnBase):
    # TODO: synthesize tall barlines for piano (multi-staff parts)

    def __post_init__(self):
        # this column should not be grown much, really
        self.flex_grow = 0.1

    def add_barline(self, barline: Glyph):
        self.add_glyph(barline)

    def _position_glyphs(self):
        for barline in self.glyphs:
            sl = self.get_staff_of_glyph(barline)

            barline.space.transform = sl.staff_coordinate_system.get_transform(
                pitch_position=0, # centered on the staff
                time_position=self.time_position
            )


def synthesize_barlines_column(
    staves: List[StaffVisual],
    rng: random.Random,
    glyph_synthesizer: GlyphSynthesizer
) -> BarlinesColumn:
    column = BarlinesColumn(staves, rng.random())

    for staff in staves:
        barline = glyph_synthesizer.synthesize_glyph_at(
            label=SmuflLabels.barlineSingle.value,
            parent_space=staff.space,
            point=Point(0, 0) # glyph positioned later
        )
        barline.space.parent_space = staff.space
        column.add_barline(barline)

    return column



from pathlib import Path
from typing import TypeVar, Type, Dict
from .AssetBundle import AssetBundle, BundleResolver
from ..config import MC_ASSETS_CACHE


T = TypeVar("T", bound=AssetBundle)


class AssetRepository(BundleResolver):
    """
    A folder on the local machine, where asset bundles are downloaded to
    and prepared. It also acts as the manager that installs and resolves
    asset bundles from this folder.
    """
    
    def __init__(self, path: Path):
        if path.exists() and not path.is_dir():
            raise Exception(
                f"The path at {path} has to be a directory or be non-existing."
            )
        
        self.path = path
        "Path to the asset repository folder"

        self.path.mkdir(parents=True, exist_ok=True)

        self._bundle_cache: Dict[Type[T], T] = dict()
        "Caches bundle instance to speed up their resolution"

    @staticmethod
    def default() -> "AssetRepository":
        """Builds a new instance of the default asset repository
        to use for this process"""
        return AssetRepository(
            Path(MC_ASSETS_CACHE).resolve()
        )
    
    def resolve_bundle(self, bundle_type: Type[T], force_install=False) -> T:
        """Ensures that a bundle is installed and returns its instance"""
        # try resolving from cache
        if bundle_type in self._bundle_cache:
            return self._bundle_cache[bundle_type]
        
        # instantiate the bundle
        bundle = bundle_type(
            bundle_directory=self.path / bundle_type.__name__,
            dependency_resolver=self
        )

        # cache the bundle instance
        self._bundle_cache[bundle_type] = bundle

        # run __post_init__ to resolve dependencies and do additional set-ups
        if hasattr(bundle, "__post_init__"):
            bundle.__post_init__()

        # do nothing if already installed
        if not bundle.needs_installation() and not force_install:
            return bundle
        
        print(f"[Smashcima Assets]: Installing bundle {bundle_type.__name__}...")

        # clear its directory
        bundle.remove()
        bundle.bundle_directory.mkdir()

        # run the installation
        bundle.install()

        # store metadata
        bundle.write_metadata()

        print(f"[Smashcima Assets]: Bundle {bundle_type.__name__} installed.")

        return bundle



from .MzkQuiltingPaperSynthesizer import MzkQuiltingPaperSynthesizer
from .NaiveStafflinesSynthesizer import NaiveStafflinesSynthesizer
from .PaperSynthesizer import PaperSynthesizer
from .SimplePageSynthesizer import SimplePageSynthesizer
from .SolidColorPaperSynthesizer import SolidColorPaperSynthesizer
from .StafflinesSynthesizer import StafflinesSynthesizer



from typing import Dict, List
from smashcima.geometry.Point import Point
from muscima.io import CropObject


class PointCloud:
    """
    Stores detected points in MUSCIMA++ crop objects for one annotated page.
    These points can then be retrieved later when finding origins and
    placement distributions.

    It's a map from objid to a list of points in the mask pixel coordinates.
    """

    def __init__(self):
        self._map: Dict[int, List[Point]] = dict()
    
    def set_points(self, o: CropObject, points: List[Point]):
        self._map[o.objid] = list(points)
    
    def get_points(self, o: CropObject) -> List[Point]:
        if o.objid not in self._map:
            raise KeyError("There are no points for object " + str(o.uid))
        return self._map[o.objid]



from typing import TypeVar

import cv2
import numpy as np

from .Contours import Contours
from .Point import Point
from .Polygon import Polygon
from .Quad import Quad
from .Vector2 import Vector2

T = TypeVar("T")


class Transform:
    """
    2D affine transform.

    When used in scene objects, it maps from the local space
    to the parent object's space.
    """

    # Based on:
    # https://www.w3.org/TR/SVGTiny12/coords.html#TransformAttribute

    def __init__(self, matrix: np.ndarray):
        assert matrix.shape == (2, 3)
        assert matrix.dtype == np.float64
        
        self.matrix = matrix
        "The 2x3 transformation matrix as a numpy array"

    @property
    def matrix3(self) -> np.ndarray:
        """The 3x3 extended matrix of this transformation"""
        return np.concatenate(
            (self.matrix, np.array([[0, 0, 1]], dtype=np.float64)),
            axis=0
        )

    @property
    def matrix2(self) -> np.ndarray:
        """The 2x2 matrix that ignores translation"""
        return self.matrix[0:2, 0:2]

    @property
    def determinant(self) -> float:
        """Returns the determinant of the affine transformation"""
        return np.linalg.det(self.matrix2)

    def inverse(self) -> "Transform":
        """Returns the inverted affine transform"""
        return Transform(cv2.invertAffineTransform(self.matrix))

    def apply_to(self, other: T) -> T:
        """Transform a vector or another transformation"""
        m = self.matrix3
        if isinstance(other, Transform):
            o = other.matrix3
            r = m.dot(o)
            return Transform(r[0:2, :]) # type: ignore
        elif isinstance(other, Vector2):
            o = np.array([[other.x], [other.y], [1]], dtype=np.float64)
            r = m.dot(o)
            return Vector2(r[0, 0], r[1, 0]) # type: ignore
        elif isinstance(other, Point):
            v = self.apply_to(other.vector)
            return Point(v.x, v.y) # type: ignore
        elif isinstance(other, Quad):
            pts = [self.apply_to(p) for p in other.points]
            return Quad(*pts) # type: ignore
        elif isinstance(other, Polygon):
            pts = [self.apply_to(p) for p in other.points]
            return Polygon(pts) # type: ignore
        elif isinstance(other, Contours):
            ps = [self.apply_to(p) for p in other.polygons]
            return Contours(ps) # type: ignore
        else:
            raise ValueError(
                "Transform applied to an unexpected type: " +
                str(type(other))
            )

    def __matmul__(self, other: T) -> T:
        return self.apply_to(other)

    def then(self, other: "Transform") -> "Transform":
        """
        Lets you chain transformations, thereby defining a new transformation.
        This method practically just inverts the matmul multiplication order
        to be chronological in terms of going from local spaces up towards
        the global scene space.
        """
        return other @ self

    @staticmethod
    def identity() -> "Transform":
        """Returns the identity transform"""
        return Transform(
            np.array([
                [1, 0, 0],
                [0, 1, 0]
            ], dtype=np.float64)
        )

    @staticmethod
    def translate(offset: Vector2) -> "Transform":
        """Returns a translation transform"""
        return Transform(
            np.array([
                [1, 0, offset.x],
                [0, 1, offset.y]
            ], dtype=np.float64)
        )

    @staticmethod
    def scale(scale: float) -> "Transform":
        """Returns a scaling transform"""
        return Transform(
            np.array([
                [scale, 0, 0],
                [0, scale, 0]
            ], dtype=np.float64)
        )

    @staticmethod
    def rotateDegCC(angle: float):
        """Creates a rotation transform for a coutner-clockwise rotation
        of a given number of degrees"""
        matrix = cv2.getRotationMatrix2D((0, 0), angle, 1)
        return Transform(matrix)



from smashcima.geometry.Rectangle import Rectangle
from ..AssetBundle import AssetBundle
from ..download_file import download_file
from pathlib import Path
from tqdm import tqdm
from dataclasses import dataclass
from typing import List, Sequence
import numpy as np
import csv
import cv2


# TODO: pick more patches and cluster them up in a zip file
# that will be downloaded from github, with MZK only for budle development
# (so that the user does not get flagged for scraping MZK)


@dataclass
class Patch:
    is_commented: bool
    mzk_uuid: str
    rectangle: Rectangle
    dpi: int

    @staticmethod
    def from_record(record: Sequence[str]) -> "Patch":
        assert len(record) == 7, "Patch record should have 7 columns"
        return Patch(
            is_commented=(record[0] == "#"),
            mzk_uuid=record[1],
            rectangle=Rectangle(
                x=int(record[2]),
                y=int(record[3]),
                width=int(record[4]),
                height=int(record[5]),
            ),
            dpi=int(record[6])
        )


class MzkPaperPatches(AssetBundle):
    def __post_init__(self):
        pass

    def install(self):
        print("Downloading MZK paper patches...")
        
        index = self.load_patch_index()
        for patch in tqdm(index):
            uuid = patch.mzk_uuid
            x = patch.rectangle.x
            y = patch.rectangle.y
            width = patch.rectangle.width
            height = patch.rectangle.height
            
            url = f"https://kramerius.mzk.cz/search/iiif/uuid:" \
                + f"{uuid}/{x},{y},{width},{height}/max/0/default.jpg"
            
            path = self.get_patch_path(patch)
            path.parent.mkdir(exist_ok=True, parents=True)
            
            download_file(url, path, with_progress_bar=False)
    
    @staticmethod
    def load_patch_index(skip_commented: bool = True) -> List[Patch]:
        """Loads the list of patches from the CSV file"""
        index_path = Path(__file__).parent / "mzk_paper_patches.csv"
        
        index: List[Patch] = []
        with open(index_path, "r") as f:
            reader = csv.reader(f, delimiter=",", quotechar='"')
            
            # header
            assert next(reader) == [
                "commented", "uuid", "x", "y", "width", "height", "dpi"
            ]
            
            # records
            for record in reader:
                patch = Patch.from_record(record)
                if skip_commented and patch.is_commented:
                    continue
                index.append(patch)
        
        return index
    
    def get_patch_path(self, patch: Patch) -> Path:
        """Returns path to the image file of the given patch"""
        uuid = patch.mzk_uuid
        x = patch.rectangle.x
        y = patch.rectangle.y
        w = patch.rectangle.width
        h = patch.rectangle.height
        dpi = patch.dpi
        filename = f"{uuid}_{x}_{y}_{w}_{h}_{dpi}.jpg"
        return self.bundle_directory / "patches" / filename

    def load_bitmap_for_patch(self, patch: Patch) -> np.ndarray:
        """Loads the patch image into the BGRA uint8 format used by
        Smashcima sprites"""
        path = self.get_patch_path(patch)

        # BGR only
        img = cv2.imread(str(path), cv2.IMREAD_COLOR)
        assert len(img.shape) == 3 # H, W, C
        assert img.shape[2] == 3 # B, G, R

        # add a non-transparent alpha channel
        alpha = np.full(
            shape=(img.shape[0], img.shape[1]),
            fill_value=255,
            dtype=np.uint8
        )
        img = np.stack(
            [img[:,:,0], img[:,:,1], img[:,:,2], alpha],
            axis=2
        )
        assert len(img.shape) == 3 # H, W, C
        assert img.shape[2] == 4 # B, G, R, A
        assert img.dtype == np.uint8

        return img



from smashcima.scene.AffineSpace import AffineSpace
from smashcima.scene.Sprite import Sprite
from smashcima.geometry.Rectangle import Rectangle
from smashcima.geometry.Transform import Transform
from smashcima.geometry.Point import Point
from smashcima.geometry.Vector2 import Vector2
from .PaperSynthesizer import PaperSynthesizer
from ..style.MzkPaperStyleDomain import MzkPaperStyleDomain
from smashcima.assets.AssetRepository import AssetRepository
from smashcima.assets.textures.MzkPaperPatches import MzkPaperPatches
from smashcima.geometry.units import mm_to_px
from .Quilter import Quilter
import numpy as np
import random
import cv2


class MzkQuiltingPaperSynthesizer(PaperSynthesizer):
    def __init__(
        self,
        assets: AssetRepository,
        style_domain: MzkPaperStyleDomain,
        rng: random.Random
    ):
        self.bundle = assets.resolve_bundle(MzkPaperPatches)
        self.style_domain = style_domain
        self.quilter = Quilter(rng)

    def synthesize_paper(
        self,
        page_space: AffineSpace,
        placement: Rectangle
    ):
        patch = self.style_domain.current_patch
        source_texture = self.bundle.load_bitmap_for_patch(patch)

        # TODO: smooth out any brightness changes over the source texture
        # (more likely this should be part of the asset bundle, not here)
        
        quilted_texture = self.quilter.quilt_texture_to_dimensions(
            source_texture=source_texture,
            target_width_px=int(mm_to_px(placement.width, patch.dpi)),
            target_height_px=int(mm_to_px(placement.height, patch.dpi)),
        )

        # create the sprite scene object
        Sprite(
            space=page_space,
            bitmap=quilted_texture,
            bitmap_origin=Point(0, 0),
            dpi=patch.dpi,
            transform=Transform.translate(Vector2(placement.x, placement.y))
        )



import random


def random_between(
    from_value: float,
    to_value: float,
    rng: random.Random
) -> float:
    """Returns a random float between two values"""
    return from_value + rng.random() * (to_value - from_value)



from smashcima.geometry.Point import Point
from smashcima.scene.semantic.Clef import Clef
from smashcima.scene.semantic.Score import Score
from smashcima.scene.visual.StaffVisual import StaffVisual
from smashcima.scene.Glyph import Glyph
from smashcima.scene.SmuflLabels import SmuflLabels
from smashcima.synthesis.GlyphSynthesizer import GlyphSynthesizer
from .ColumnBase import ColumnBase
from typing import List, Set
import random


class ClefsColumn(ColumnBase):
    def __post_init__(self) -> None:
        self.clefs: List[Clef] = []

        # this column should not be grown much, really
        self.flex_grow = 0.1

    def add_clef(self, clef: Clef, clef_glyph: Glyph):
        self.clefs.append(clef)
        self.add_glyph(clef_glyph)

    def _position_glyphs(self):
        for clef, clef_glyph in zip(self.clefs, self.glyphs):
            sl = self.get_staff_of_glyph(clef_glyph)

            pitch_position = Clef.clef_line_to_pitch_position(clef.line)

            clef_glyph.space.transform = sl.staff_coordinate_system.get_transform(
                pitch_position=pitch_position,
                time_position=self.time_position
            )


def synthesize_header_clefs(
    staves: List[StaffVisual],
    rng: random.Random,
    glyph_synthesizer: GlyphSynthesizer,
    score: Score,
    measure_index: int
) -> ClefsColumn:
    column = ClefsColumn(staves, rng.random())

    # verification logic
    handled_staves: Set[int] = set()

    # go through all the parts
    for part in score.parts:
        measure = part.measures[measure_index]
        event = measure.first_event

        # and all the clefs in the part
        # (applying to the first event of the measure we are rendering)
        for staff_number, clef in event.attributes.clefs.items():

            # get the proper stafflines instance
            staff_index = score.first_staff_index_of_part(part) \
                + (staff_number - 1)
            staff = staves[staff_index]

            # determine the glyph class
            label = SmuflLabels.clef_from_clef_sign(
                clef_sign=clef.sign,
                small=False
            ).value

            # synthesize the glyph
            glyph = glyph_synthesizer.synthesize_glyph_at(
                label=label,
                parent_space=staff.space,
                point=Point(0, 0) # glyph positioned later
            )
            column.add_clef(clef, glyph)

            # verification logic
            if staff_index in handled_staves:
                raise Exception(
                    f"The staff {staff_index} had a clef be created twice!"
                )
            handled_staves.add(staff_index)
    
    # verify there is one clef for each staff
    assert len(handled_staves) == len(staves), \
        "All staves must get a clef when synthesizing the header clefs."
    
    return column


def synthesize_change_clefs() -> ClefsColumn:
    # TODO: synthesize clef changes (small clefs)
    pass



from dataclasses import dataclass
from typing import List, Optional

from ..geometry.Transform import Transform
from .SceneObject import SceneObject


@dataclass
class AffineSpace(SceneObject):
    """
    Affine spaces define the visual hierarchy of the scene. Only visual elements
    require their presence. Their hierarchy allows for defining local spatial
    relationships in local coordinate systems. 2D affine transforms are used
    to traverse the affine space hierarchy.
    """

    parent_space: Optional["AffineSpace"] = None
    "What other space does this space belong under"

    transform: Transform = Transform.identity()
    """Transform that translates from this space's local coordinates
    to the parent's space coordinates, effectively defining the placement of
    this space within the parent space."""

    def get_children(self) -> List["AffineSpace"]:
        """Returns child affine spaces as a list"""
        return AffineSpace.many_of(self, lambda s: s.parent_space)
    
    def get_root(self) -> "AffineSpace":
        """Returns the top-most root affine space"""
        space = self
        while space.parent_space is not None:
            space = space.parent_space
        return space

    def transform_from(self, sub_space: "AffineSpace") -> Transform:
        """Returns the transform from the given space to the current space"""
        if sub_space is None:
            raise ValueError("The given space should not be None")
        
        t = Transform.identity()
        s: Optional[AffineSpace] = sub_space
        
        while s is not None:
            if s is self:
                return t
            
            t = t.then(s.transform)
            s = s.parent_space

        raise Exception("The given sub space is not attached under this space")



from enum import Enum


class SmashcimaLabels(str, Enum):
    """Glyph and region classification labels that are not in SMuFL"""
    
    # Staff Glyphs
    #
    # While SMuFL defines staves, they are indended for text-based applications
    # only. Therefore we define our own classes here, since their semantics
    # are slightly different. But the class names are analogous.
    # https://www.w3.org/2019/03/smufl13/tables/staves.html
    staffLine = "smashcima::staffLine"
    ledgerLine = "smashcima::ledgerLine"

    # Line Glyphs
    #
    # Line glyphs are not really represented in SMuFL,
    # so instead they must be added to the set of classes here.
    beam = "smashcima::beam"
    beamHook = "smashcima::beamHook"
    slur = "smashcima::slur"

    # Layout Regions
    staff = "smashcima::staff"
    grandStaff = "smashcima::grandStaff"
    system = "smashcima::system"
    staffMeasure = "smashcima::staffMeasure"
    grandStaffMeasure = "smashcima::grandStaffMeasure"
    systemMeasure = "smashcima::systemMeasure"

    # Isolated Flags
    #
    # IMPORTANT: For flags as such use the SMuFL class.
    # These are reserved for cases when the flag is split up into individual
    # isolated flag stokes. For example MUSCIMA++ does this.
    isolatedFlag8thUp = "smashcima::flag8thUp"
    isolatedFlag8thDown = "smashcima::flag8thDown"
    isolatedFlag16thUp = "smashcima::flag16thUp"
    isolatedFlag16thDown = "smashcima::flag16thDown"
    isolatedFlag32ndUp = "smashcima::flag32ndUp"
    isolatedFlag32ndDown = "smashcima::flag32ndDown"
    isolatedFlag64thUp = "smashcima::flag64thUp"
    isolatedFlag64thDown = "smashcima::flag64thDown"
    isolatedFlag128thUp = "smashcima::flag128thUp"
    isolatedFlag128thDown = "smashcima::flag128thDown"
    isolatedFlag256thUp = "smashcima::flag256thUp"
    isolatedFlag256thDown = "smashcima::flag256thDown"
    isolatedFlag512thUp = "smashcima::flag512thUp"
    isolatedFlag512thDown = "smashcima::flag512thDown"
    isolatedFlag1024thUp = "smashcima::flag1024thUp"
    isolatedFlag1024thDown = "smashcima::flag1024thDown"



from enum import Enum


class StemValue(str, Enum):
    """Orientation of a stem for a note"""
    # https://www.w3.org/2021/06/musicxml40/musicxml-reference/data-types/stem-value/
    down = "down"
    up = "up"
    none = "none"

    # double = "double"
    # We assign stem orientations to chords and a chord cannot have a double
    # stem orientation. Only a notehead does. Therefore this value does not
    # make sense in our context. Should the value be encountered in MusicXML,
    # the parser must deal with it somehow. Stem synthesizers assume a chord
    # has only one well determined stem orientation.


from typing import Dict, List, Set, Tuple

from smashcima.scene.Glyph import Glyph

from .LineList import LineList
from .MppGlyphMetadata import MppGlyphMetadata
from .PackedGlyph import PackedGlyph


class SymbolRepository:
    """
    Extracted glyphs from MUSCIMA++ so that they can be sampled
    by the synthesizer
    """
    def __init__(self) -> None:
        self.all_glyphs: List[PackedGlyph] = []
        "Contains all glyphs in the repository in one big list"

        self.all_writers: Set[int] = set()
        "Contains all writer numbers that can be sampled from this repository"

        self.glyphs_by_class: Dict[str, List[PackedGlyph]] = {}
        "Contains all glyphs grouped by glyph class"

        self.glyphs_by_class_and_writer: Dict[Tuple[str, int], List[PackedGlyph]] = {}
        "Contains all glyphs grouped by glyph class and MPP writer number"

    def add_glyphs(self, glyphs: List[Glyph]):
        for glyph in glyphs:
            self.add_glyph(glyph)

    def add_glyph(self, glyph: Glyph):
        metadata = MppGlyphMetadata.of_glyph(glyph)
        assert metadata is not None, "Glyph must have MPP metadata attached"
        
        packed_glyph = PackedGlyph.pack(glyph)

        self.all_glyphs.append(packed_glyph)

        self.all_writers.add(metadata.mpp_writer)

        key1 = glyph.label
        self.glyphs_by_class.setdefault(key1, [])
        self.glyphs_by_class[key1].append(packed_glyph)

        key2 = (glyph.label, metadata.mpp_writer)
        self.glyphs_by_class_and_writer.setdefault(key2, [])
        self.glyphs_by_class_and_writer[key2].append(packed_glyph)
    
    def remove_writer(self, writer: int):
        """Removes a given writer from the repository completely.
        Can be used to remove testing set when performing synthesis."""
        self.all_glyphs = list(filter(
            lambda pg: pg.mpp_writer == writer,
            self.all_glyphs
        ))
        self.all_writers.remove(writer)
        for key1 in self.glyphs_by_class.keys():
            self.glyphs_by_class[key1] = list(filter(
                lambda pg: pg.mpp_writer == writer,
                self.glyphs_by_class[key1]
            ))
        for key2 in self.glyphs_by_class_and_writer.keys():
            self.glyphs_by_class_and_writer[key2] = list(filter(
                lambda pg: pg.mpp_writer == writer,
                self.glyphs_by_class_and_writer[key2]
            ))

    def index_lines(self, labels: List[str]):
        """Build line lookup index for each lines collection"""
        for label in labels:
            if label in self.glyphs_by_class:
                self.glyphs_by_class[label] = \
                    LineList(self.glyphs_by_class[label])

            for writer in self.all_writers:                
                if (label, writer) in self.glyphs_by_class_and_writer:
                    self.glyphs_by_class_and_writer[label, writer] = \
                        LineList(self.glyphs_by_class_and_writer[label, writer])



from .SceneObject import SceneObject
from .AffineSpace import AffineSpace
from .Sprite import Sprite
from ..geometry.Point import Point
from ..geometry.Rectangle import Rectangle
from dataclasses import dataclass


@dataclass
class ScenePoint(SceneObject):
    """Like the geometric point, but is placed within the scene visual hierarchy,
    therefore belonging under an affine space."""
    
    point: Point
    "The point, in the coordinate space of the attached affine space"

    space: AffineSpace
    "The affine space the point is situated in"

    def transform_to(self, other_space: AffineSpace) -> Point:
        """Returns the point, as if it was placed in a higher affine space"""
        return other_space.transform_from(self.space).apply_to(self.point)

    def detach(self):
        """Detaches the scene point from the scene hierarchy"""
        self.space = None
    
    def place_debug_overlay(self, radius=0.25, color=(0, 0, 255, 128)) -> Sprite:
        """Places a sprite that acts as debugging overlay for the point"""
        return Sprite.debug_box(
            space=self.space,
            rectangle=Rectangle(
                x=self.point.x - radius,
                y=self.point.y - radius,
                width=radius * 2,
                height=radius * 2
            ),
            fill_color=color,
            border_width=0
        )



from ...AssetBundle import AssetBundle
from ...datasets.MuscimaPP import MuscimaPP
from smashcima.exporting.DebugGlyphRenderer import DebugGlyphRenderer
from .MppPage import MppPage
from smashcima.scene.SmuflLabels import SmuflLabels
from smashcima.scene.SmashcimaLabels import SmashcimaLabels
from .get_symbols import \
    get_full_noteheads, \
    get_empty_noteheads, \
    get_normal_barlines, \
    get_whole_rests, \
    get_half_rests, \
    get_quarter_rests, \
    get_eighth_rests, \
    get_sixteenth_rests, \
    get_g_clefs, \
    get_f_clefs, \
    get_c_clefs, \
    get_stems, \
    get_beams, \
    get_beam_hooks, \
    get_ledger_lines, \
    get_flags, \
    get_duration_dots, \
    get_staccato_dots, \
    get_accidentals, \
    get_brackets_and_braces, \
    get_time_marks
from .SymbolRepository import SymbolRepository
from .MppGlyphMetadata import MppGlyphMetadata
from pathlib import Path
import pickle
from tqdm import tqdm
import shutil
import cv2
from typing import Optional


# Re-install asset bunle during development by running:
# .venv/bin/python3 -m smashcima.assets.glyphs.muscima_pp --debug

class MuscimaPPGlyphs(AssetBundle):
    def __post_init__(self) -> None:
        self._symbol_repository_cache: Optional[SymbolRepository] = None

        self.muscima_pp = self.dependency_resolver.resolve_bundle(MuscimaPP)

    @property
    def symbol_repository_path(self) -> Path:
        return self.bundle_directory / "symbol_repository.pkl"
    
    def install(self):
        """Extracts data from the MUSCIMA++ dataset and bundles it up
        in the symbol repository in a pickle file."""
        document_paths = list(
            self.muscima_pp.cropobjects_directory.glob("CVC-MUSCIMA_*-ideal.xml")
        )

        repository = SymbolRepository()

        # go through all the MUSCIMA++ XML files
        for document_path in tqdm(document_paths):
            page = MppPage.load(document_path)

            # and extract glyphs
            repository.add_glyphs(get_full_noteheads(page))
            repository.add_glyphs(get_empty_noteheads(page))
            repository.add_glyphs(get_normal_barlines(page))
            repository.add_glyphs(get_whole_rests(page))
            repository.add_glyphs(get_half_rests(page))
            repository.add_glyphs(get_quarter_rests(page))
            repository.add_glyphs(get_eighth_rests(page))
            repository.add_glyphs(get_sixteenth_rests(page))
            repository.add_glyphs(get_g_clefs(page))
            repository.add_glyphs(get_f_clefs(page))
            repository.add_glyphs(get_c_clefs(page))
            repository.add_glyphs(get_stems(page))
            repository.add_glyphs(get_beams(page))
            repository.add_glyphs(get_beam_hooks(page))
            repository.add_glyphs(get_ledger_lines(page))
            repository.add_glyphs(get_duration_dots(page))
            repository.add_glyphs(get_staccato_dots(page))
            repository.add_glyphs(get_accidentals(page))
            repository.add_glyphs(get_brackets_and_braces(page))
            repository.add_glyphs(get_time_marks(page))
            # (flags must come after stems)
            glyphs_8th_flag, glyphs_16th_flag = get_flags(page)
            repository.add_glyphs(glyphs_8th_flag)
            repository.add_glyphs(glyphs_16th_flag)

            # and construct line indexes
            repository.index_lines([
                SmuflLabels.stem.value,
                SmashcimaLabels.beam.value,
                SmashcimaLabels.beamHook.value,
                SmashcimaLabels.ledgerLine.value
            ])

            # TODO: and extract distributions

        # dump the repository into a pickle file
        with open(self.symbol_repository_path, "wb") as file:
            pickle.dump(repository, file)
            print("Writing...", self.symbol_repository_path)
    
    def load_symbol_repository(self) -> SymbolRepository:
        """Loads the symbol repository from its pickle file"""
        if self._symbol_repository_cache is None:
            with open(self.symbol_repository_path, "rb") as file:
                repository = pickle.load(file)
            assert isinstance(repository, SymbolRepository)
            self._symbol_repository_cache = repository
        
        return self._symbol_repository_cache
    
    def build_debug_folder(self):
        """Creates a debug folder in the bundle folder, where it dumps
        all the extracted glyphs for visual inspection."""
        repository = self.load_symbol_repository()
        
        debug_folder = self.bundle_directory / "debug"
        shutil.rmtree(debug_folder, ignore_errors=True)
        debug_folder.mkdir()

        # glyphs
        glyph_renderer = DebugGlyphRenderer()
        for label, glyphs in repository.glyphs_by_class.items():
            glyphs_folder = debug_folder / label.replace(":", "-")
            glyphs_folder.mkdir()

            print(label, "...")
            for glyph in tqdm(glyphs):
                meta = MppGlyphMetadata.of_glyph(glyph)
                cv2.imwrite(
                    str(glyphs_folder / (meta.mpp_crop_object_uid + ".png")),
                    glyph_renderer.render(glyph)
                )



from dataclasses import dataclass
from .RestSemantic import RestSemantic
from .TypeDuration import TypeDuration
from .Pitch import Pitch
from fractions import Fraction
from typing import Optional


@dataclass
class MeasureRest(RestSemantic):
    def __init__(
        self,
        fractional_duration: Fraction,
        display_pitch: Optional[Pitch]
    ):
        """Constructs the measure rest, given the fractional duration of
        the whole measure (the number of beats)"""
        super().__init__(
            type_duration=TypeDuration.whole,
            augmentation_dots=0,
            fractional_duration=fractional_duration,
            display_pitch=display_pitch
        )



from dataclasses import dataclass
from typing import Optional

from smashcima.scene import Glyph, SceneObject

from .MppPage import MppPage


@dataclass
class MppGlyphMetadata(SceneObject):
    """Metadata information object for a glyph from the MUSCIMA++ dataset"""

    glyph: Glyph
    "The glyph that this metadata refers to"
    
    mpp_writer: int
    "Writer index (1 to 50) from the MUSCIMA++ dataset"

    mpp_piece: int
    "Music piece index (1 to 20) from the MUSCIMA++ dataset"

    mpp_numeric_objid: int
    "Id number assigned to the corresponding crop object in the MUSCIMA++ dataset"

    def __post_init__(self):
        assert self.glyph is not None
        assert self.mpp_writer is not None
        assert self.mpp_writer >= 1 and self.mpp_writer <= 50
        assert self.mpp_piece is not None
        assert self.mpp_piece >= 1 and self.mpp_piece <= 20
    
    @property
    def mpp_crop_object_uid(self) -> str:
        """The full XML id of the corresponding crop object"""
        assert self.mpp_numeric_objid is not None
        w = str(self.mpp_writer).zfill(2)
        n = str(self.mpp_piece).zfill(2)
        i = str(self.mpp_numeric_objid)
        return f"MUSCIMA-pp_1.0___CVC-MUSCIMA_W-{w}_N-{n}_D-ideal___{i}"

    @staticmethod
    def stamp_glyph(glyph: Glyph, mpp_page: MppPage, numeric_objid: int):
        # just create an instance and that's it
        # the glyph's inlinks will hold on to the instance
        MppGlyphMetadata(
            glyph=glyph,
            mpp_writer=mpp_page.mpp_writer,
            mpp_piece=mpp_page.mpp_piece,
            mpp_numeric_objid=numeric_objid
        )
    
    @classmethod
    def of_glyph(cls, glyph: Glyph):
        return cls.of(glyph, lambda m: m.glyph)

    @classmethod
    def of_glyph_or_none(cls, glyph: Optional[Glyph]):
        return cls.of_or_none(glyph, lambda m: m.glyph)



from ..AssetBundle import AssetBundle
from ..download_file import download_file
import zipfile
from pathlib import Path


DOWNLOAD_URL = "https://lindat.mff.cuni.cz/repository/xmlui/bitstream/" + \
    "handle/11372/LRT-2372/MUSCIMA-pp_v1.0.zip"


class MuscimaPP(AssetBundle):
    def install(self):
        print("Downloading MUSCIMA++ dataset...")
        downloaded_zip = self.bundle_directory / "MUSCIMA-pp_v1.0.zip"
        download_file(
            DOWNLOAD_URL,
            downloaded_zip
        )

        print("Extracting the zip...")
        with zipfile.ZipFile(downloaded_zip, "r") as zip_ref:
            zip_ref.extractall(self.bundle_directory)
        downloaded_zip.unlink()

        print("Checking bundle directory structure...")
        assert (self.bundle_directory / "v1.0").is_dir()
        assert (self.bundle_directory / "v1.0" / "data").is_dir()
        assert (
            self.bundle_directory / "v1.0" / "data" / "cropobjects_withstaff"
        ).is_dir()
        assert (self.cropobjects_directory).is_dir()
    
    @property
    def cropobjects_directory(self) -> Path:
        """Returns path to the directory with all cropobject XMLs"""
        return self.bundle_directory / "v1.0" / "data" / "cropobjects_withstaff"



from dataclasses import dataclass
from enum import Enum


class TimeSymbol(str, Enum):
    """How to render a time signature"""
    # https://www.w3.org/2021/06/musicxml40/musicxml-reference/data-types/time-symbol/
    common = "common" # the "C" symbol
    cut = "cut" # the "slashed C" symbol
    normal = "normal" # two numbers atop each other


@dataclass
class TimeSignature:
    """Represents music notation time signature"""
    # https://www.w3.org/2021/06/musicxml40/musicxml-reference/elements/time/
    
    beats: int
    "How many beats are there in one measure. It's the top number."

    beat_type: int
    "What note type is one beat. It's the bottom number, 4 = quarter."

    symbol: TimeSymbol
    "How to render the time signature"

    def __post_init__(self):
        assert self.beats in list(range(1, 16+1)), \
            "Unexpected beat number per measure (expected 1-16)"
        assert self.beat_type in [2, 4, 8, 16], "Unexpected beat type."



from dataclasses import dataclass, field
from ..SceneObject import SceneObject
from typing import List, Optional
from .Note import Note
from .Event import Event
from .StemValue import StemValue
from .TypeDuration import TypeDuration


@dataclass
class Chord(SceneObject):
    """Chord is a collection of notes that share one stem. Even a single note
    has a chord, since it needs to encode its stem information here. Musically
    a chord is a group of notes with the same onset and duration though.
    A whole note chord is also a chord, despite not having a stem - there the
    stem has value "none". (this is for consistency)"""
    
    notes: List[Note] = field(default_factory=list)
    "Links to all notes within this chord"

    stem_value: Optional[StemValue] = None
    """What orientation does the stem have.
    If StemValue.none, the stem is missing (e.g. whole note).
    If None, the stem orientation is unknown
    and must be inferred during rendering."""

    @classmethod
    def of_note(cls, note: Note):
        return cls.of(note, lambda c: c.notes)
    
    def get_event(self) -> Event:
        """Returns the event that contains notes in this chord"""
        if len(self.notes) == 0:
            raise Exception("The chord is empty, there are no notes.")
        return Event.of_durable(self.notes[0])
    
    def get_type_duration(self) -> TypeDuration:
        """Returns the type duration of all notes in this chord"""
        if len(self.notes) == 0:
            raise Exception("The chord is empty, there are no notes.")
        return self.notes[0].type_duration
    
    def add_note(self, note: Note, stem_value: Optional[StemValue] = None):
        """Adds a note into the chord with a stem value"""
        # store stem value if missing
        if self.stem_value is None:
            self.stem_value = stem_value

        # validate stem value if present
        if self.stem_value is not None and stem_value is not None:
            assert self.stem_value == stem_value, \
                "All notes in a chord must have the same stem value"
        
        # validate that all the notes share the same event (have the same onset)
        if len(self.notes) > 0:
            event = Event.of_durable(self.notes[0])
            inserted_event = Event.of_durable(note)
            assert event is inserted_event, \
                "All notes in a chord must be in the same event (same onset)"
        
        # validate that all the notes have the same type duration
        if len(self.notes) > 0:
            type_duration = self.notes[0].type_duration
            assert type_duration == note.type_duration, \
                "All notes in a chord must have the same type duration"
        
        # update the list of notes
        notes = [*self.notes, note]
        notes.sort(key=lambda n: n.pitch.get_linear_pitch()) # ascending by pitch
        self.notes = notes



from .AccidentalValue import AccidentalValue
from .AttributesChange import AttributesChange
from .Attributes import Attributes
from .BeamedGroup import BeamedGroup
from .BeamValue import BeamValue
from .Chord import Chord
from .Clef import Clef
from .ClefSign import ClefSign
from .Durable import Durable
from .Event import Event
from .KeySignature import KeySignature
from .Measure import Measure
from .MeasureRest import MeasureRest
from .Note import Note
from .Part import Part
from .Pitch import Pitch
from .RestSemantic import RestSemantic
from .ScoreEvent import ScoreEvent
from .ScoreMeasure import ScoreMeasure
from .Score import Score
from .StaffSemantic import StaffSemantic
from .StemValue import StemValue
from .TimeSignature import TimeSignature
from .TypeDuration import TypeDuration



from smashcima.scene.semantic.Score import Score
from smashcima.scene.semantic.ScoreEvent import ScoreEvent
from smashcima.scene.semantic.Note import Note
from smashcima.scene.visual.StaffVisual import StaffVisual
from smashcima.scene.visual.Accidental import Accidental
from smashcima.scene.visual.Notehead import Notehead
from smashcima.scene.SmuflLabels import SmuflLabels
from smashcima.synthesis.GlyphSynthesizer import GlyphSynthesizer
from smashcima.geometry.Point import Point
from smashcima.random_between import random_between
from .ColumnBase import ColumnBase
from .Skyline import Skyline
from typing import List


class AccidentalsColumn(ColumnBase):
    def __post_init__(self) -> None:
        self.accidentals: List[Accidental] = []
    
    def add_accidental(self, accidental: Accidental):
        assert accidental.notehead is not None
        self.glyphs.append(accidental.glyph)
        self.accidentals.append(accidental)

    def _position_glyphs(self):
        self.place_accidentals()

    def place_accidentals(self):
        for staff in self.staves:
            self.place_accidentals_for_staff(staff)

    def place_accidentals_for_staff(self, staff: StaffVisual):
        # extract all accidentals in this staff
        accidentals: List[Accidental] = []
        for accidental in self.accidentals:
            if self.get_staff_of_glyph(accidental.notehead.glyph) is not staff:
                continue
            accidentals.append(accidental)
        
        # extract all noteheads in this staff
        # TODO: this is ugly, it's crawling the graph in ways it should not
        noteheads: List[Notehead] = []
        for glyph in self.glyphs:
            notehead = Notehead.of_or_none(glyph, lambda n: n.glyph)
            if notehead is None:
                continue
            if self.get_staff_of_glyph(glyph) is not staff:
                continue
            noteheads.append(notehead)

        # constants
        SPACING = random_between(0.2, 1.0, self.rng) # between accidentals
        
        # where is the column in the staff space coords
        origin_x = staff.staff_coordinate_system.get_transform(
            0, self.time_position
        ).apply_to(Point(0, 0)).x

        # define the skyline by noteheads
        # (in the column-local space increasing to the right)
        skyline = Skyline(ground_level=0)
        for notehead in noteheads:
            bbox_global = notehead.glyph.get_bbox_in_space(staff.space)
            skyline.overlay_box(
                minimum=bbox_global.top,
                maximum=bbox_global.bottom,
                level=(origin_x - bbox_global.left)
            )

        # place accidentals
        for accidental in accidentals:
            assert accidental.notehead in noteheads, \
                "Not all noteheads have been extracted to build the skyline base"

            bbox_global = accidental.glyph.get_bbox_in_space(staff.space)
            bbox_local = accidental.glyph.get_bbox_in_space(accidental.glyph.space)

            skyline_left = skyline.drop_box(
                minimum=bbox_global.top,
                maximum=bbox_global.bottom,
                thickness=bbox_global.width + SPACING
            )
            accidental.glyph.space.transform = staff.staff_coordinate_system \
                .get_transform(
                    pitch_position=accidental.notehead.pitch_position,
                    time_position=(
                        self.time_position - bbox_local.left - skyline_left
                    )
                )

def synthesize_accidentals_column(
    column: AccidentalsColumn,
    staves: List[StaffVisual],
    glyph_synthesizer: GlyphSynthesizer,
    score: Score,
    score_event: ScoreEvent
):
    for event in score_event.events:
        for durable in event.durables:
            if not isinstance(durable, Note):
                continue
            note = durable

            # skip notes with no accidental
            if note.accidental_value is None:
                continue

            # create accidental
            staff_index = score.staff_index_of_durable(note)
            glyph = glyph_synthesizer.synthesize_glyph_at(
                label=SmuflLabels.accidental_from_accidental_value(
                    note.accidental_value
                ).value,
                parent_space=staves[staff_index].space,
                point=Point(0, 0) # glyph positioned later
            )
            accidental = Accidental(
                glyph=glyph,
                notehead=Notehead.of_note(note)
            )
            column.add_accidental(accidental)



from smashcima.scene.AffineSpace import AffineSpace
from smashcima.scene.Sprite import Sprite
from smashcima.geometry.Rectangle import Rectangle
from smashcima.geometry.Transform import Transform
from smashcima.geometry.Point import Point
from smashcima.geometry.Vector2 import Vector2
from .PaperSynthesizer import PaperSynthesizer
from smashcima.geometry.units import mm_to_px
from typing import Tuple
import numpy as np


COLOR_WHITE = (255, 255, 255, 255)
COLOR_TRANSPARENT = (0, 0, 0, 0)


class SolidColorPaperSynthesizer(PaperSynthesizer):
    def __init__(self):
        self.color: Tuple[int, int, int, int] = COLOR_WHITE
        "Color to use as the paper, BGRA uint8 format"

        self.dpi = 72
        "The DPI at which to rasterize the sprite"

    def synthesize_paper(
        self,
        page_space: AffineSpace,
        placement: Rectangle
    ):
        # width_px
        img = np.empty(
            shape=(
                int(mm_to_px(placement.height, self.dpi)),
                int(mm_to_px(placement.width, self.dpi)),
                4
            ),
            dtype=np.uint8
        )
        img[:,:] = self.color
        
        # create the sprite scene object
        Sprite(
            space=page_space,
            bitmap=img,
            bitmap_origin=Point(0, 0),
            dpi=self.dpi,
            transform=Transform.translate(Vector2(placement.x, placement.y))
        )



from dataclasses import dataclass, field
from ..SceneObject import SceneObject
from typing import List, Optional, Dict, Tuple, Iterator
from .Chord import Chord
from .BeamValue import BeamValue


@dataclass
class BeamedGroup(SceneObject):
    """A group of notes that share at least one beam. Rests cannot be present
    in a beamed group, because rests lack stems and beam attaches to stems.
    Although a rest can be in the middle of a beamed group optically, it is not
    present here semantically."""
    
    chords: List[Chord] = field(default_factory=list)
    """Chords in this beamed group, in chronological order
    (chord is a group of notes that share a stem)"""

    beam_values: List[Dict[int, BeamValue]] = field(default_factory=list)
    """For each chord there is a dictionary, mapping beam numbers to beam values"""
    
    @classmethod
    def of_chord_or_none(cls, chord: Optional[Chord]):
        return cls.of_or_none(chord, lambda g: g.chords)

    def add_chord(self, chord: Chord, beam_values: Dict[int, BeamValue]):
        """Adds a new chord into the beamed group"""
        assert chord not in self.chords, \
            "Cannot add a chord twice into a beamed group."
        
        self.chords = [*self.chords, chord]
        self.beam_values.append(beam_values)
    
    @property
    def is_complete(self) -> bool:
        """Returns true if the beamed group has been terminated on the last chord."""
        if len(self.chords) == 0:
            return False
        return self.beam_values[-1].get(1, None) == BeamValue.end
    
    def iterate_beams(self) -> Iterator[Tuple[int, List[Chord]]]:
        """Return all beams to be drawn"""
        open_beams: Dict[int, List[Chord]] = dict()
        for chord, beam_values in zip(self.chords, self.beam_values):
            for beam_number, beam_value in beam_values.items():
                if beam_value == BeamValue.begin:
                    open_beams[beam_number] = [chord]
                elif beam_value == BeamValue.continue_beam:
                    open_beams[beam_number].append(chord)
                elif beam_value == BeamValue.end:
                    open_beams[beam_number].append(chord)
                    chords = open_beams[beam_number]
                    del open_beams[beam_number]
                    yield (beam_number, chords)
        assert len(open_beams) == 0
    
    def iterate_hooks(self) -> Iterator[Tuple[int, Chord, BeamValue]]:
        """Return all hooks to be drawn"""
        for chord, beam_values in zip(self.chords, self.beam_values):
            for beam_number, beam_value in beam_values.items():
                if beam_value in [BeamValue.forward_hook, BeamValue.backward_hook]:
                    yield (beam_number, chord, beam_value)



from dataclasses import dataclass

from ..AffineSpace import AffineSpace
from ..Glyph import Glyph
from ..SceneObject import SceneObject
from ..SmashcimaLabels import SmashcimaLabels
from .StaffCoordinateSystem import StaffCoordinateSystem


@dataclass
class StaffVisual(SceneObject):
    """Represents the visual stafflines for a staff on the page"""

    width: float
    "Width (in millimeters) of these stafflines"

    staff_coordinate_system: StaffCoordinateSystem
    "Coordinate system that maps from pitch-time space to 2D position"

    space: AffineSpace
    """The affine space that contains all glyphs (noteheads, rests, accidentals)
    on these stafflines"""

    glyph: Glyph
    """Glyph of the stafflines themselves (all lines together, usually composite
    where each line is a sub-glyph). Child of the staff space."""

    staff_height: float
    """Average height of all stafflines in millimeters"""

    def __post_init__(self):
        assert self.glyph.region.label == SmashcimaLabels.staff.value, \
            "StaffVisual must have 'smashcima::staff' as its glyph label."



from ._version import __version__

# -----------------------------------------------------------------------------
# import important types that should also be accessible at the top level
from .geometry import *
from .scene import *
from .orchestration.Model import Model
from .synthesis.GlyphSynthesizer import GlyphSynthesizer
from .synthesis.LineSynthesizer import LineSynthesizer

# -----------------------------------------------------------------------------
# import sub-modules to make them accessible from this module
# TODO: assets
from smashcima import exporting
from smashcima import geometry
# smashcima.jupyter must always be imported explicitly, since it depends
# on jupyter, which is an optional dependency
from smashcima import loading
from smashcima import orchestration
from smashcima import scene
from smashcima import synthesis
from smashcima import config



from ..Glyph import Glyph
from ..SceneObject import SceneObject
from .Notehead import Notehead
from .RestVisual import RestVisual
from dataclasses import dataclass
from typing import List, Union


@dataclass
class AugmentationDot(SceneObject):
    """Glyph of an augmentation dot"""

    glyph: Glyph
    """The dot glyph"""

    owners: List[Union[Notehead, RestVisual]]
    """The glyph(s) that is(are) affected by this augmentation dot.
    For noteheads in dense chords, augmentation dots may be shared.
    Otherwise it's usually one-to-one"""

    augmentation_dot_index: int
    """Which dot (out of many for the owner) is this dot?
    Starting from 1, increasing 2, 3, 4."""

    pitch_position: int
    """At what pitch position should the augmentation dot be placed."""

    def detach(self):
        """Unlink the glyph from the scene"""
        self.glyph.detach()
        self.owners = []



import xml.etree.ElementTree as ET
from ..scene.semantic.Score import Score
from ..scene.semantic.Part import Part
from ..scene.semantic.Measure import Measure
from ..scene.semantic.Note import Note
from ..scene.semantic.RestSemantic import RestSemantic
from ..scene.semantic.MeasureRest import MeasureRest
from ..scene.semantic.TypeDuration import TypeDuration
from ..scene.semantic.Pitch import Pitch
from ..scene.semantic.AttributesChange import AttributesChange
from ..scene.semantic.KeySignature import KeySignature
from ..scene.semantic.Clef import Clef
from ..scene.semantic.ClefSign import ClefSign
from ..scene.semantic.TimeSignature import TimeSignature, TimeSymbol
from ..scene.semantic.Chord import Chord
from ..scene.semantic.StemValue import StemValue
from ..scene.semantic.BeamedGroup import BeamedGroup
from ..scene.semantic.BeamValue import BeamValue
from ..scene.semantic.AccidentalValue import AccidentalValue
from typing import List, TextIO, Optional, Dict, Union
from fractions import Fraction
from dataclasses import dataclass, field
from pathlib import Path
import io


@dataclass
class _ScoreState:
    new_system_measures: List[Measure] = field(default_factory=list)
    "Measures that should be placed on new systems (line breaks)"

    new_page_measures: List[Measure] = field(default_factory=list)
    "Measures that should be placed on new pages (page breaks)"


@dataclass
class _PartState:
    part_id: str
    "MusicXML ID of the currently parsed part, used for error localization"

    part: Part
    "The part that is being constructed as it's being parsed"

    measure_number: Optional[str] = None
    "Currently parsed measure number, or None if it was not yet defined/parsed"

    divisions: Optional[int] = None
    "Number of MusicXML duration units per quarter note"

    beats_per_measure: Optional[int] = None
    "How many beats there are in one measure"

    beat_type: Optional[int] = None
    "Type of the beat (inverse of the beat fractional duration)"

    beat_fractional_duration: Optional[Fraction] = None
    "How many quarter notes does a single beat take up"

    measure_fractional_duration: Optional[Fraction] = None
    "How many quarter notes does a whole measure take up"


@dataclass
class _MeasureState:
    part_state: _PartState
    "Reference to the part state"

    measure: Measure
    "The measure that is being constructed as it's being parsed"

    last_chord: Optional[Chord] = None
    """Last chord whose note was constructed. If we hit the <chord/> element,
    then this chord should be extended. Otherwise a new one should be created."""

    fractional_measure_onset: Fraction = Fraction(0, 1) # zero
    """How many quarter notes from the start of the measure are we currently"""

    current_beamed_group: Optional[BeamedGroup] = None
    "The beamed group that is currently being constructed."

    def seek_forward(self, fractional_duration: Fraction):
        self.fractional_measure_onset += fractional_duration
        mfd = self.part_state.measure_fractional_duration
        if self.fractional_measure_onset > mfd:
            self.fractional_measure_onset = mfd
    
    def seek_backup(self, fractional_duration: Fraction):
        self.fractional_measure_onset -= fractional_duration
        if self.fractional_measure_onset < 0:
            self.fractional_measure_onset = Fraction(0, 1) # zero


IGNORED_MEASURE_ELEMENTS = set([
    # https://www.w3.org/2021/06/musicxml40/musicxml-reference/elements/measure-partwise/
    "direction", "harmony", "figured-bass", "sound", "listening",
    "grouping", "link", "bookmark", "barline"
])


class MusicXmlLoader:
    """Loads MusicXML into the scene data model"""

    def __init__(self, errout: Optional[TextIO] = None):
        self._errout = errout or io.StringIO()
        "Print errors and warnings here"

        self._score_state: Optional[_ScoreState] = None
        "Score state, not None when parsing a score"
        
        self._part_state: Optional[_PartState] = None
        "Part state, not None when parsing a part"

        self._measure_state: Optional[_MeasureState] = None
        "Measure state, not None when parsing a measure"

    def _error(self, *values):
        if self._part_state:
            p = self._part_state.part_id
            m = self._part_state.measure_number
            header = f"[ERROR][P:{p} M:{m}]:"
        else:
            header = f"[ERROR]:"
        print(header, *values, file=self._errout)

    def load_file(self, path: Union[Path, str]) -> Score:
        """Loads a score from a MusicXML file"""
        # TODO: handle .mxl files as well
        with open(str(path), "r") as file:
            return self.load(ET.parse(file))

    def load_xml(self, xml: str) -> Score:
        fake_file = io.StringIO(xml)
        return self.load(ET.parse(fake_file))

    def load(self, tree: ET.ElementTree) -> Score:
        """Loads a score from a MusicXML XML tree"""
        score_partwise_element = tree.getroot()
        if score_partwise_element.tag != "score-partwise":
            raise Exception("The loader expects the <score-partwose> " + \
                            "tag to be the root of the file.")
        
        score = self._load_score_partwise(score_partwise_element)
        score.validate()
        return score
    
    def _load_score_partwise(self, score_partwise_element: ET.Element) -> Score:
        assert score_partwise_element.tag == "score-partwise"

        self._score_state = _ScoreState()

        parts: List[Part] = []

        # go through all the parts
        part_list_element = score_partwise_element.find("part-list")
        for score_part_element in part_list_element.findall("score-part"):
            part_id = score_part_element.attrib.get("id")
            if part_id is None:
                raise Exception("<score-part> element is missing an ID.")

            # find the part element
            part_elements = [
                p for p in score_partwise_element.findall("part")
                if p.attrib.get("id") == part_id
            ]
            if len(part_elements) == 0:
                raise Exception(f"Cannot find <part> with ID '{part_id}'.")
            part_element = part_elements[0]

            # and parse it
            part = self._load_part(score_part_element, part_element, part_id)
            parts.append(part)

        # create the part instance
        score = Score(
            parts=parts,
            new_system_measure_indices=set(
                Part.of_measure(m).measures.index(m)
                for m in self._score_state.new_system_measures
            ),
            new_page_measure_indices=set(
                Part.of_measure(m).measures.index(m)
                for m in self._score_state.new_page_measures
            )
        )

        self._score_state = None

        return score
    
    def _load_part(
        self,
        score_part_element: ET.Element,
        part_element: ET.Element,
        part_id: str
    ) -> Part:
        assert score_part_element.tag == "score-part" # part header
        assert part_element.tag == "part" # part measures
        assert score_part_element.attrib["id"] == part_id
        assert part_element.attrib["id"] == part_id

        self._part_state = _PartState(
            part_id=part_id,
            part=Part()
        )

        for measure_element in part_element:
            measure = self._load_measure(measure_element)
            self._part_state.part.append_measure(measure)
        
        part = self._part_state.part
        self._part_state = None

        # set attributes for each event in the part
        part.compute_event_attributes()
        
        return part
    
    def _load_measure(self, measure_element: ET.Element) -> Measure:
        assert measure_element.tag == "measure"

        self._part_state.measure_number = measure_element.attrib.get("number")
        self._measure_state = _MeasureState(
            part_state=self._part_state,
            measure=Measure()
        )

        for element in measure_element:
            if element.tag in IGNORED_MEASURE_ELEMENTS:
                continue
            elif element.tag == "note":
                self._load_note(element)
            elif element.tag == "attributes":
                self._load_attributes(element)
            elif element.tag == "backup":
                self._load_backup(element)
            elif element.tag == "forward":
                self._load_forward(element)
            elif element.tag == "print":
                self._load_print(element)
            else:
                self._error(
                    "Unexpected <measure> element:",
                    element,
                    element.attrib
                )
        
        measure = self._measure_state.measure
        self._measure_state = None
        
        return measure

    def _load_note(self, note_element: ET.Element):
        assert note_element.tag == "note"

        # <grace>
        def _grace():
            grace_element = note_element.find("grace")
            is_grace_note = grace_element is not None
            is_grace_slash = False
            if is_grace_note:
                if grace_element.attrib.get("slash") == "yes":
                    is_grace_slash = True
            return is_grace_note, is_grace_slash
        
        # <chord>
        def _chord():
            chord_element = note_element.find("chord")
            is_chord = chord_element is not None
            return is_chord

        # <rest> or <pitch>
        def _rest_or_pitch():
            rest_element = note_element.find("rest")
            is_measure_rest = False
            pitch: Optional[Pitch] = None
            if rest_element is not None:
                is_measure_rest = rest_element.attrib.get("measure") == "yes"
            else:
                pitch_element = note_element.find("pitch")
                step = pitch_element.find("step").text
                octave = pitch_element.find("octave").text
                alter = None
                alter_element = pitch_element.find("alter")
                if alter_element is not None:
                    alter = alter_element.text
                pitch = Pitch.parse(octave, step, alter)
            return pitch, is_measure_rest
        
        # <display-step> and <display-octave> for rests
        def _display_pitch() -> Optional[Pitch]:
            rest_element = note_element.find("rest")
            if rest_element is None:
                return None
            step_element = rest_element.find("display-step")
            octave_element = rest_element.find("display-octave")
            if step_element is None or octave_element is None:
                return None
            return Pitch.parse(
                octave=octave_element.text,
                step=step_element.text
            )
        
        # <voice>
        def _voice():
            voice_element = note_element.find("voice")
            voice_name: Optional[str] = None
            if voice_element is not None:
                voice_name = voice_element.text
            return voice_name

        # <type>, missing only for rest measures
        def _type(is_measure_rest: bool):
            type_element = note_element.find("type")
            type_duration: Optional[TypeDuration] = None
            if type_element is not None:
                type_duration = TypeDuration(type_element.text)
            elif is_measure_rest:
                pass # leave type duration at None
            else:
                self._error("Note does not have <type>:", ET.tostring(note_element))
            return type_duration
        
        # <time-modification> (tuplets rhythm-wise)
        def _time_modification() -> Optional[Fraction]:
            # triplet is 3in2 (3 actual in 2 normal notes)
            # which translates to 2/3 = 0.66 times the normal duration
            if note_element.find("time-modification") is None:
                return None # no time modification
            actual = note_element.find("time-modification/actual-notes").text
            normal = note_element.find("time-modification/normal-notes").text
            return Fraction(int(normal), int(actual))
        
        # <dot>
        def _dot() -> int:
            return len(note_element.findall("dot"))
        
        # <accidental>
        def _accidental() -> Optional[AccidentalValue]:
            if note_element.find("accidental") is None:
                return None # no accidental
            return AccidentalValue(note_element.find("accidental").text)
        
        # <stem>
        def _stem() -> Optional[StemValue]:
            stem_element = note_element.find("stem")
            if stem_element is None:
                return None # stem orientation is unknown
            if stem_element.text not in ["up", "down", "double", "none"]:
                self._error(
                    f"Unknown stem type '{stem_element.text}'.",
                    ET.tostring(note)
                )
            return StemValue(stem_element.text)

        # <staff>
        def _staff() -> int:
            staff_element = note_element.find("staff")
            if staff_element is None:
                return 1
            staff = int(staff_element.text)
            assert staff >= 1, "Staff number must be 1 or more"
            assert staff <= self._part_state.part.staff_count, \
                "Staff number must not exceed the defined staff count"
            return staff

        # <beam>
        def _beam() -> Dict[int, BeamValue]:
            values = {}
            for element in note_element.findall("beam"):
                v = BeamValue(element.text)
                n = int(element.attrib.get("number"))
                values[n] = v
            return values

        # TODO: <tied>
        # TODO: <tuplet>
        
        # === decode the <note> element ===

        is_grace_note, is_grace_slash = _grace()
        is_chord = _chord()
        pitch, is_measure_rest = _rest_or_pitch() # None pitch for rests
        display_pitch = _display_pitch() # for rests only
        voice_name = _voice()
        type_duration = _type(is_measure_rest) # None for measure rests
        time_modification = _time_modification()
        duration_dots = _dot()
        accidental_value = _accidental()
        stem_value = _stem()
        staff_number = _staff()
        beam_values = _beam()
        
        fractional_duration = self._part_state.measure_fractional_duration \
            if is_measure_rest \
            else self._decode_fractional_duration(
                type_duration=type_duration,
                duration_dots=duration_dots,
                time_modification=time_modification
            )

        # right now just ignore grace notes
        # TODO: decode grace notes
        if is_grace_note:
            return None
        
        # determine onset and advance onset state
        onset = self._measure_state.fractional_measure_onset
        if is_chord:
            onset -= fractional_duration
        if not is_chord:
            self._measure_state.seek_forward(fractional_duration)
        
        # determine the current chord instance
        current_chord: Optional[Chord] = None
        if is_measure_rest or pitch is None:
            self._measure_state.last_chord = None
        else:
            if is_chord: # extend old chord
                current_chord = self._measure_state.last_chord
            else: # start a new chord
                current_chord = Chord()
        self._measure_state.last_chord = current_chord # update state

        # load beam information
        # (only the first note in a chord has this information,
        # others don't repeat it, unlike with stems)
        if len(beam_values) > 0:
            if self._measure_state.current_beamed_group is None:
                self._measure_state.current_beamed_group = BeamedGroup()
            self._measure_state.current_beamed_group.add_chord(
                current_chord, beam_values
            )
            if self._measure_state.current_beamed_group.is_complete:
                self._measure_state.current_beamed_group = None

        # handle measure rests
        if is_measure_rest:
            assert self._part_state.measure_fractional_duration is not None, \
                "Measure duration has not been loaded, yet we're parsing notes"
            assert onset == 0, "Measure rest should begin at 0 onset"
            self._measure_state.measure.add_durable(
                durable=MeasureRest(
                    fractional_duration=fractional_duration,
                    display_pitch=display_pitch
                ),
                onset=onset,
                staff_number=staff_number,
            )
            return
        
        # handle rests
        if pitch is None:
            self._measure_state.measure.add_durable(
                durable=RestSemantic(
                    type_duration=type_duration,
                    augmentation_dots=duration_dots,
                    fractional_duration=fractional_duration,
                    display_pitch=display_pitch
                ),
                onset=onset,
                staff_number=staff_number
            )
            return
        
        # handle notes
        note = Note(
            pitch=pitch,
            type_duration=type_duration,
            augmentation_dots=duration_dots,
            fractional_duration=fractional_duration,
            accidental_value=accidental_value,
        )
        self._measure_state.measure.add_durable(
            durable=note,
            onset=onset,
            staff_number=staff_number
        )

        current_chord.add_note(note, stem_value)
    
    def _decode_fractional_duration(
        self,
        type_duration: TypeDuration,
        duration_dots: int,
        time_modification: Optional[Fraction]
    ) -> Fraction:
        # first get the number of quarter notes
        fractional_duration = type_duration.to_quarter_multiple()

        # now apply duration dots
        dot_duration = fractional_duration / 2
        for _ in range(duration_dots):
            fractional_duration += dot_duration
            dot_duration /= 2
        
        # finally apply time modification
        if time_modification is not None:
            fractional_duration *= time_modification
        
        return fractional_duration

    def _load_attributes(self, attributes_element: ET.Element):
        # https://www.w3.org/2021/06/musicxml40/musicxml-reference/elements/attributes/
        assert attributes_element.tag == "attributes"

        assert self._part_state is not None, \
            "Part state must be initialized before loading <attributes>"
        assert self._measure_state is not None, \
            "Measure state must be initialized before loading <attributes>"
        
        change = AttributesChange()

        # divisions:
        divisions_element = attributes_element.find("divisions")
        if divisions_element is not None:
            self._load_divisions(divisions_element)
        
        # staves
        staves_element = attributes_element.find("staves")
        if staves_element is not None:
            self._part_state.part.staff_count = int(staves_element.text)
        
        # key signature
        key_element = attributes_element.find("key")
        if key_element is not None:
            staff_number = key_element.attrib.get("number")
            fifths = int(key_element.find("fifths").text)
            if staff_number is None:
                for i in range(1, self._part_state.part.staff_count + 1):
                    change.keys[i] = KeySignature(fifths=fifths)
            else:
                change.keys[int(staff_number)] = KeySignature(fifths=fifths)

        # time signature
        time_element = attributes_element.find("time")
        if time_element is not None:
            change.time_signature = self._load_time_signature(time_element)

        # clef
        clef_elements = attributes_element.findall("clef")
        for clef_element in clef_elements:
            staff_number = int(clef_element.attrib.get("number", "1"))
            after_barline = (clef_element.attrib.get("after-barline", "no") == "yes")
            sign = ClefSign(clef_element.find("sign").text.upper())
            line = int(clef_element.find("line").text)
            change.clefs[staff_number] = Clef(
                sign=sign,
                line=line,
                after_barline=after_barline
            )

        # add the attributes change into the measure
        self._measure_state.measure.add_attributes_change(
            change=change,
            onset=self._measure_state.fractional_measure_onset
        )
    
    def _load_divisions(self, divisions_element: ET.Element):
        assert divisions_element.tag == "divisions"

        self._part_state.divisions = int(divisions_element.text)

        assert self._part_state.divisions > 0, \
            "<divisions> should be a positive number"
    
    def _load_time_signature(self, time_element: ET.Element) -> TimeSignature:
        assert time_element.tag == "time"
        
        # parse XML tree
        beats_element = time_element.find("beats")
        assert beats_element is not None, \
            "<beats> must be present in <time> element"
        beat_type_element = time_element.find("beat-type")
        assert beat_type_element is not None, \
            "<beat-type> must be present in <time> element"
        
        # extract time signature values
        beats_per_measure = int(beats_element.text) # typically 4, or 3
        assert beats_per_measure > 0
        beat_type = int(beat_type_element.text) # typically 4
        assert beat_type > 0

        # extract time symbol
        symbol = TimeSymbol(
            time_element.attrib.get("symbol", "normal")
        )

        # compute
        beat_fractional_duration = Fraction(1, beat_type) / Fraction(1, 4)
        measure_fractional_duration = beat_fractional_duration * beats_per_measure

        # modify part state
        self._part_state.beats_per_measure = beats_per_measure
        self._part_state.beat_type = beat_type
        self._part_state.beat_fractional_duration = beat_fractional_duration
        self._part_state.measure_fractional_duration = measure_fractional_duration

        return TimeSignature(
            beats=beats_per_measure,
            beat_type=beat_type,
            symbol=symbol
        )
    
    def _load_backup(self, backup_element: ET.Element):
        assert backup_element.tag == "backup"

        # get the duration
        backup_duration = int(backup_element.find("duration").text)
        assert backup_duration > 0

        # number of quarter notes
        fractional_duration = Fraction(
            backup_duration,
            self._part_state.divisions
        )

        # modify measure state
        self._measure_state.seek_backup(fractional_duration)
    
    def _load_forward(self, forward_element: ET.Element):
        assert forward_element.tag == "forward"

        # get the duration
        forward_duration = int(forward_element.find("duration").text)
        assert forward_duration > 0

        # number of quarter notes
        fractional_duration = Fraction(
            forward_duration,
            self._part_state.divisions
        )

        # modify measure state
        self._measure_state.seek_forward(fractional_duration)
    
    def _load_print(self, print_element: ET.Element):
        assert print_element.tag == "print"

        # parse line breaks
        if print_element.attrib.get("new-system") == "yes":
            self._score_state.new_system_measures.append(
                self._measure_state.measure
            )
        
        # parse page breaks
        if print_element.attrib.get("new-page") == "yes":
            self._score_state.new_page_measures.append(
                self._measure_state.measure
            )



from ...geometry.Vector2 import Vector2
from ...scene.AffineSpace import AffineSpace
from ...scene.visual.StaffVisual import StaffVisual
import abc


class StafflinesSynthesizer(abc.ABC):
    """Interface for a stafflines synthesizer"""

    @property
    @abc.abstractmethod
    def staff_height(self) -> float:
        """Returns the average height of synthesized staff in millimeters"""
        raise NotImplementedError

    @abc.abstractmethod
    def synthesize_stafflines(
        self,
        page_space: AffineSpace,
        position: Vector2,
        width: float,
    ) -> StaffVisual:
        """Synthesizes a new StaffVisual object.

        :param page_space: Stafflines will be placed into this space.
        :param position: Position of the left endpoint of the middle staffline.
        :param width: Width of the staff in millimeters.
        """
        raise NotImplementedError



from ..Glyph import Glyph
from ..SceneObject import SceneObject
from .Stem import Stem
from dataclasses import dataclass


@dataclass
class Flag(SceneObject):
    """Glyph of a flag with flag-related properties"""

    glyph: Glyph
    "Glyph of the flag"

    stem: Stem
    "The stem that the flag is attached to"

    def detach(self):
        """Unlink the glyph from the scene"""
        self.glyph.detach()
        self.stem = None



from dataclasses import dataclass, field
from typing import Optional, Dict
from .Clef import Clef
from .TimeSignature import TimeSignature
from .KeySignature import KeySignature


@dataclass
class AttributesChange:
    """A change in the part attributes (clefs, keys), attached to an event"""
    # https://www.w3.org/2021/06/musicxml40/musicxml-reference/elements/attributes/

    clefs: Dict[int, Clef] = field(default_factory=dict)
    """Clefs by staff number (staff number starts from 1 for the top staff).
    If a clef does not change for the given staff, it is not present here."""

    keys: Dict[int, KeySignature] = field(default_factory=dict)
    """Key signatures by staff number (staff number starts from 1 for the top
    staff). If a key does not change for the given staff, it is not present."""

    time_signature: Optional[TimeSignature] = None
    "Time signature change, must be at the start of a measure."



from typing import Any, Callable, Type, TypeVar


T = TypeVar("T")


class _Dummy:
    """
    Dummy class: When you get any attribute, it returns that attributes name
    """
    def __getattribute__(self, name: str) -> Any:
        return name


def nameof_via_dummy(examined_type: Type[T], probe: Callable[[T], Any]) -> str:
    """Get the name of an object property via a lambda function by substituting
    a dummy instance instead of the real examined type.
    
    Can be used like this:
    nameof_via_dummy(MyType, lambda my_type: my_type.my_property)
    """
    return str(probe(_Dummy())) # type: ignore



from enum import Enum
from dataclasses import dataclass, field
from typing import Optional, Union


STEP_ORDER = ["C", "D", "E", "F", "G", "A", "B"]


class Octave(int, Enum):
    """Octave number of the scientific picth notation"""
    # https://www.w3.org/2021/06/musicxml40/musicxml-reference/data-types/octave/
    o0 = 0
    o1 = 1
    o2 = 2
    o3 = 3
    o4 = 4
    o5 = 5
    o6 = 6
    o7 = 7
    o8 = 8
    o9 = 9


class Step(str, Enum):
    """Step name of the scientific pitch notation"""
    # https://www.w3.org/2021/06/musicxml40/musicxml-reference/data-types/step/
    C = "C"
    D = "D"
    E = "E"
    F = "F"
    G = "G"
    A = "A"
    B = "B"


class Alter(int, Enum):
    """Semitone alter number (based on MusicXML)"""
    # https://www.w3.org/2021/06/musicxml40/musicxml-reference/data-types/semitones/
    flat = -1
    none = 0
    sharp = 1


@dataclass
class Pitch:
    """Represents an audible pitch in the scientific pitch notation."""
    # https://www.w3.org/2021/06/musicxml40/musicxml-reference/elements/pitch/
    octave: Octave
    step: Step
    alter: Alter = field(default=Alter.none)

    @staticmethod
    def parse(octave: str, step: str, alter: Optional[str] = None) -> "Pitch":
        return Pitch(
            octave=Octave(int(octave)),
            step=Step(step),
            alter=Alter.none if alter is None else Alter(int(alter))
        )
    
    def get_linear_pitch(self, with_alters=False) -> Union[int, float]:
        """Converts the pitch object into a number where 0 is A0 note
        and then it increases by one for each tone. If you include alters,
        the returned value is a float with +-0.5 based on the alter."""
        step_index = STEP_ORDER.index(self.step.value)
        linear_pitch = self.octave.value * len(STEP_ORDER) + step_index
        if with_alters:
            if self.alter == Alter.sharp:
                linear_pitch += 0.5
            elif self.alter == Alter.flat:
                linear_pitch -= 0.5
        return linear_pitch
    
    @staticmethod
    def from_linear_pitch(linear_pitch: int) -> "Pitch":
        """Reconstructs a pitch object from an integer linear pitch"""
        assert type(linear_pitch) is int, \
            "Float reconstructions are not supported"
        step_index = linear_pitch % len(STEP_ORDER)
        octave = linear_pitch // len(STEP_ORDER)
        return Pitch(Octave(octave), Step(STEP_ORDER[step_index]))
    
    def __repr__(self) -> str:
        suffix = ""
        if self.alter == Alter.sharp:
            suffix += "+1"
        if self.alter == Alter.flat:
            suffix += "-1"
        return f"Pitch({self.step.value}{self.octave.value}{suffix})"


if __name__ == "__main__":
    # Prints out all the pitches and their linear representation
    # (linear representation is used for staff positioning and conversions)
    print("Pitches and their linear representation:")
    for octave in range(9):
        for step in STEP_ORDER:
            pitch = Pitch.parse(octave, step)
            linear_pitch = pitch.get_linear_pitch()
            assert pitch.octave == Pitch.from_linear_pitch(linear_pitch).octave
            assert pitch.step == Pitch.from_linear_pitch(linear_pitch).step
            print(
                repr(pitch),
                str(linear_pitch).zfill(2),
                end=" | "
            )
        print()



import abc

from smashcima.geometry.Transform import Transform


class StaffCoordinateSystem(abc.ABC):
    @abc.abstractmethod
    def get_transform(
        self,
        pitch_position: float,
        time_position: float
    ) -> Transform:
        """Converts stafflines coordinates (pitch position in staffspace/2 units,
        where zero is the middle staffline, 4 is the topmost staffline;
        time position in milimeters from the begining of the staff)
        to a transform to a specific place (and possibly rotation)
        on the staff."""
        # TODO: maybe time position should be in different units than
        # millimeters, this needs to be inspected. How does staffline
        # spacing affect temporal spacing of notes. Are they related? How?
        raise NotImplementedError



import pickle
from typing import Optional

from smashcima.scene import Glyph, LineGlyph

from .MppGlyphMetadata import MppGlyphMetadata


class PackedGlyph:
    """Contains a sub-pickled glyph so that the loading of the whole repository
    is much faster. The glyph is unpacked only once sampled and needed.
    
    This is because creating python objects takes most of the time,
    so this approach reduces the number of python objects created when the
    symbol repository is unpickled from the file system.

    This trick speeds up loading the repository pickle file from about
    15 seconds down to under a second.
    """
    
    def __init__(
        self,
        line_length: Optional[float],
        mpp_writer: int,
        data: bytes
    ):
        self.line_length = line_length
        """If a line glyph, stores its length for sampling lookups"""
        
        self.mpp_writer = mpp_writer
        """Number of the MUSCIMA++ writer"""

        self.data = data
        """The pickled glyph instance"""

    @staticmethod
    def pack(glyph: Glyph) -> "PackedGlyph":
        line_length: Optional[float] = None
        if isinstance(glyph, LineGlyph):
            line_length = glyph.line_length
        return PackedGlyph(
            line_length=line_length,
            mpp_writer=MppGlyphMetadata.of_glyph(glyph).mpp_writer,
            data=pickle.dumps(glyph)
        )
    
    def unpack(self) -> Glyph:
        return pickle.loads(self.data)



from dataclasses import dataclass
from typing import Union

from ..Glyph import Glyph
from ..SceneObject import SceneObject
from ..semantic import Clef, MeasureRest, Pitch, RestSemantic, TypeDuration
from .StaffVisual import StaffVisual

# Display Pitch
# -------------
# MusicXML defines so-called display pitch for rests. By default, this value
# is missing and it's up to the rendering software to position the rest properly.
# When present, this position is overriden. The pitch depends on the active clef
# and for the percussion clef, the G2 treble is assumed.
#
# However there is little specification on how the display pitch translates
# to a position on the staff, especially for all the different glyphs of
# varous rest types.
#
# MuseScore assumes the default display pitch to be B4 (for the trebble clef,
# or rather, the pitch of the center staffline). This is the case even for
# the whole rest, which could be argued is either in the upper space (
# which is what the MXL user manual example suggests), or hangs from the
# line above that.
#
# For consistency, we decided to follow the MuseScore approach
# (because it treats all the rests the same way in the semantic domain)
# and we introduce pitch_position offsets. These offsets are a global constant
# that could be changed in case you want to interpret display pitches for
# rests differently. Alternatively, you can monkey patch the two static
# method that perform the conversion (default pitch and pitch to pitch position).
#
# Info from W3C:
# https://www.w3.org/2021/06/musicxml40/musicxml-reference/elements/display-step/
#
# Info from MXL:
# https://usermanuals.musicxml.com/MusicXML/Content/EL-MusicXML-rest.htm
# https://usermanuals.musicxml.com/MusicXML/Content/EL-MusicXML-display-step.htm


@dataclass
class RestVisual(SceneObject):
    """Glyph of a rest symbol"""

    glyph: Glyph
    "The glyph of the rest object"

    rest_semantic: Union[RestSemantic, MeasureRest, None]
    "The semantic rest that this glyph represents"

    clef: Clef
    "What clef applies to the rest"

    staff: StaffVisual
    "What stafflines is the rest placed onto"

    pitch_position: int
    "Display pitch position of the rest on the stafflines"

    def detach(self):
        """Unlink the glyph from the scene"""
        self.glyph.detach()
        self.rest_semantic = None
        self.clef = None
        self.staff = None

    @classmethod
    def of_rest(cls, rest: RestSemantic):
        return cls.of(rest, lambda r: r.rest_semantic)

    @staticmethod
    def default_display_pitch(clef: Clef, type_duration: TypeDuration) -> Pitch:
        """Reurns the default pitch position for rests in the given clef"""
        # NOTE: type_duration is ignored, this is to that an advanced user
        # can monkey patch this method and have different defaults
        # for different rest types

        # as does MuseScore, we assume the center line
        # to be the default display pitch for rests
        return clef.pitch_position_to_pitch(pitch_position=0)

    @staticmethod
    def display_pitch_to_glyph_pitch_position(
        clef: Clef,
        display_pitch: Pitch,
        type_duration: TypeDuration
    ) -> int:
        """Converts a display pitch of a rest, in a clef, of rest type given
        by type duration, into the pitch position of the rest's glyph
        (where on the staff to put the glyph's origin vertically)"""
        pitch_position = clef.pitch_to_pitch_position(display_pitch)

        # whole note is the only exception, where its origin is rendered
        # one line above the center line
        if type_duration == TypeDuration.whole:
            return pitch_position + 2 # one line up
        
        return pitch_position
